--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- Date: June 2001
-- Status: limited functionality
-- Todo: entailBCons, consistentBcons to work on NEq
--       adjust normalize to new type language
--       check entailBCons
--------------------------------------------------------

module Builtin(BCons(Eq, NEq), subst2eq, projBCons, projBConsWithUnifier, 
		  projBConsKeepIdentities,
		  td_projBCons, entailBCons, consistentBCons,buildEP,buildUP, 
		  bcIx, bcJust, bcLeft, bcRight, newBCJust, addBCJust,
		  entailBCons2, njEq, njNEq, catBCJust, remRedundantBCons,
		  bconsLocs
		  ) where

import List
import Name
import Types
import Justify
import Pretty
import Misc(E(..))
import IOExts

-------------------------------
-- Type and data definitions --
------------------------------- 

-- Built-in constraints

data BCons = Eq Type Type Just | NEq Type Type Just deriving Eq

-- builds an unjustified Eq
njEq  x y = Eq x y noJust
njNEq x y = NEq x y noJust

bcJust = bcIx
bcIx (Eq  _ _ ix) = ix
bcIx (NEq _ _ ix) = ix

bcLeft (Eq  t1 _ _) = t1
bcLeft (NEq t1 _ _) = t1
bcRight (Eq  _ t2 _) = t2
bcRight (NEq _ t2 _) = t2

newBCJust j (Eq  t1 t2 _) = Eq t1 t2 j
newBCJust j (NEq t1 t2 _) = NEq t1 t2 j

nubBCJust (Eq t1 t2 j) = Eq t1 t2 (nubJust j)
nubBCJust (NEq t1 t2 j) = NEq t1 t2 (nubJust j)

-- delete this junk!
--newBCTypesJust j (Eq t1 t2 j')  = Eq (newJustType j t1) (newJustType j t2) j'
--newBCTypesJust j (NEq t1 t2 j') = NEq (newJustType j t1) (newJustType j t2) j'

addBCJust j (Eq  t1 t2 k) = Eq t1 t2 (addJust j k)
addBCJust j (NEq t1 t2 k) = Eq t1 t2 (addJust j k)
catBCJust j (Eq  t1 t2 k) = Eq t1 t2 (catJust j k)
catBCJust j (NEq t1 t2 k) = Eq t1 t2 (catJust j k)

instance Justifiable BCons where
    unJust (Eq  _ _ j) = unJust j
    unJust (NEq _ _ j) = unJust j
    reJust j (Eq a b _) = Eq a b j
    reJust j (NEq a b _) = NEq a b j


instance PPrint BCons where

    pprint (Eq t1 t2 x) = case (t1, t2) of
	(TyCon (Name ("()",_)), TApp (TyCon (Name ("[]",_))) _)-> 
		textPD "False" -- <+> text ("_"++show x)
	
	(TyCon (Name ("()",_)), TyCon (Name ("()",_))) -> 
		textPD "True" -- <+> text ("_"++show x)
	_ -> do let dt1 = pprint t1
		dt2 <- pprint t2
		dt1 <+> text "=" <+> dt2 -- <+> text ("_"++show x)

    
    pprint (NEq t1 t2 x) = do let dt1 = pprint t1
			      dt2 <- pprint t2
			      dt1 <+> text "/=" <+> dt2
		      
  
instance Show BCons where
  show (Eq ty1 ty2 j)  = show ty1 ++ " = " ++ show ty2 -- ++ "_" ++ show j
  show (NEq ty1 ty2 _) = show ty1 ++ " != " ++ show ty2



---------------------------
-- Instance declarations --
---------------------------


instance TC BCons where
   apply s (Eq t1 t2 x)  = Eq (apply s t1) (apply s t2) x
   apply s (NEq t1 t2 x) = NEq (apply s t1) (apply s t2) x
   fv (Eq t1 t2 _)  = (fv t1) ++ (fv t2)
   fv (NEq t1 t2 _) = (fv t1) ++ (fv t2)
   occurs v (Eq t1 t2 _)  = (occurs v t1) || (occurs v t2)
   occurs v (NEq t1 t2 _) = (occurs v t1) || (occurs v t2)
   rename (t@(Eq t1 t2 _))  = do s <- renameSubst (fv (t1:[t2]))    -- note duplicate type variables
                                 return (apply s t)                -- don't cause problems!
   rename (t@(NEq t1 t2 _)) = do s <- renameSubst (fv (t1:[t2]))
                                 return (apply s t)
   format (Eq t1 t2 _)  = (format t1) ++ "=" ++ (format t2)
   format (NEq t1 t2 _) = (format t1) ++ "/=" ++ (format t2)
 

---------------
-- Functions --
---------------

-- represent substition as set of equality constraints
-- projected on a set of variables
-- result = projv tvs. s
subst2eq :: Subst -> [TVar] -> [BCons]
subst2eq s tvs = foldl f [] tvs
     where
       f eq y = let t=lookupVar s y
		    t' = trace ("lookup " ++ show y ++ ": " ++ show t) t
                in case t of
                     Var y' -> if (notElem y' tvs) || (y==y') then eq   -- get rid of identities
                               else (Eq (Var y) t noJust):eq
                     _      -> (Eq (Var y) t noJust):eq

-- as above, but keeps identity equations
subst2eqKeepIdentities :: Subst -> [TVar] -> [BCons]
subst2eqKeepIdentities s tvs = foldl f [] tvs
     where
       f eq y = let t = lookupVar s y
                in  (Eq (Var y) t noJust):eq


-- projection of type variables, plus simplification (not optimal see below!)
-- assumption: bc is consistent
-- note that unify needs to take into account the projected variables!!!
projBCons :: [TVar] -> [BCons] -> [BCons]
projBCons _ [] = []
projBCons vs bc =
    let (eq, neq) = foldl f ([],[]) bc   -- split into equality and inequality
        f (eq,neq) (Eq t1 t2 x)   = ((Eq t1 t2 x):eq,neq)
        f (eq, neq) (NEq t1 t2 x) = (eq, (NEq t1 t2 x):neq)
        eq' = buildUP eq
        Succ mgu = unify vs idSubst eq'     -- we assume bc is consistent
        -- note, mgu = eq 
        -- proj vs. mgu = { Eq v mgu(v) | v in vs }
	mgu' = trace ("mgu2: " ++ show mgu) mgu
        eq'' = subst2eq mgu vs                         -- projected equality constraints
        neq' = apply mgu neq
        -- proj vs. neq' = proj vs.{ NEq t1 t1 |  NEq t1 t2 in neq', and 
        --                           if t1 (or t2) = Var v then v in vs }
        -- note, not optimal,
        -- example:  proj b. {NEq (a->a) (Int->a) }
        neq'' = foldl h [] neq'
        h neq (NEq t1 t2 x) = case t1 of
                                Var v1 -> if elem v1 vs then ((NEq t1 t2 x):neq)
                                               else neq
                                _      -> case t2 of
                                           Var v2 -> if elem v2 vs then ((NEq t1 t2 x):neq)
                                                     else neq 
    in neq'' ++ eq'' 


-- projection of type variables, plus simplification (not optimal see below!)
-- assumption: bc is consistent
-- note that unify needs to take into account the projected variables!!!
projBConsKeepIdentities :: [TVar] -> [BCons] -> [BCons]
projBConsKeepIdentities _ [] = []
projBConsKeepIdentities vs eq =
    let eq' = buildUP eq
        Succ mgu = unify vs idSubst eq'      -- we assume bc is consistent
        eq'' = subst2eqKeepIdentities mgu vs -- projected equality constraints
    in eq'' 



-- same as projBCons, except takes a unifier as an additional argument, which
-- it uses instead of generating a new one
projBConsWithUnifier :: Subst -> [TVar] -> [BCons] -> [BCons]
projBConsWithUnifier _   _  [] = []
projBConsWithUnifier mgu vs bc =
    let (eq, neq) = foldl f ([],[]) bc
        f (eq,neq) (Eq t1 t2 x)  = ((Eq t1 t2 x):eq,neq)
        f (eq,neq) (NEq t1 t2 x) = (eq, (NEq t1 t2 x):neq)
        eq'' = subst2eq mgu vs
{-        
	neq' = apply mgu neq
        neq'' = foldl h [] neq'
        h neq (NEq t1 t2) 
	    = case t1 of
                Var v1 -> if elem v1 vs then ((NEq t1 t2):neq)
					else neq
                _      -> case t2 of
                            Var v2 -> if elem v2 vs then ((NEq t1 t2):neq)
                                                    else neq 
-}						    
    in {- neq'' ++ -} eq''

td_projBCons :: Subst -> [TVar] -> [BCons] -> [BCons]
td_projBCons _   _  [] = []
td_projBCons mgu vs bc =
    let (eq, neq) = foldl f ([],[]) bc
        f (eq, neq) (Eq  t1 t2 x) = ((Eq t1 t2 x):eq,neq)
        f (eq, neq) (NEq t1 t2 x) = (eq, (NEq t1 t2 x):neq)
        eq'' = subst2eq' mgu vs
    in  eq''

    where   subst2eq' s tvs = 
		let f eq y = let t = lookupVar s y
			     in  (Eq (Var y) t (locToJust (varLoc y))):eq
		in  foldl f [] tvs

-- calculates the proper set of variables on which we project on
calculate vs subst = let res = map (\ a -> lookup a subst) vs
                         res' = zip res vs
                         res'' = [ (case x of
                                     Just t -> fv t
                                     Nothing -> [v])    | (x,v) <- res']
                         res''' = concat res''
                   in res'''

-- entailment test among builtin constraints
--  bc1 implies proj tvs.bc2
-- assumption we only consider equality constraints for the moment
-- constraints are consistent
-- bc1 implies proj tvs. bc2  iff
--    proj tvs. phi(bc2) == True where phi mgu of bc1
entailBCons :: [BCons] -> [TVar] -> [BCons] -> Bool
entailBCons bc1 tvs bc2 = 
      let eqp = buildUP bc1
          Succ mgu = unify tvs idSubst eqp      -- we assume constraints are consistent
          bc = apply mgu bc2
          tvs' = calculate tvs mgu              -- projected variables might have changed
	  p   = projBCons tvs' bc
	  res = null p
      in  {- trace ("\n tvs: " ++ show tvs ++
                 "\n tvs': " ++ show tvs' ++
                 "\n bc1: " ++ show bc1 ++
                 "\n bc2: " ++ show bc2 ++
                 "\n bc: " ++ show bc ++
                 "\n p: " ++ show p) -} res

entailBCons2 bc1 bc2 =
      let eqp = buildUP bc1
          Succ mgu = unify [] idSubst eqp
          bc = apply mgu bc2
      in and (map (\ x -> case x of Eq t1 t2 _ -> t1 == t2) bc)
      -- and [] => True
	 -- all (\(Eq t1 t2 _)-> t1 == t2) bc


-- consistency test for builtin constraints
-- assumption we only consider equality constraints for the moment
consistentBCons :: [BCons] -> Bool
consistentBCons bc =
      let eqp = buildUP bc
      in case (unify [] idSubst eqp) of
           Succ _ -> True
           Err _  -> False



-- testversions for handling of equality and inequality constraints


{- adjust to new type language

normalize :: [BCons] -> E [[(Int,Int)]]
normalize xs = case (normalize1 xs) of
                Succ xs' -> Succ (normalize2 xs')
                Err _    -> Err ""

-- we transform inequations (NEq t1 t2) into a list of a pair of Int's
-- the component of the pair correspond to the head type constructor of t1, t2
-- lists are sorted, and redundancies such as
-- (NEq (f 0) (g 1)) reduced to [] and
-- (NEq (g 0 1) (g 2 3)) reduced to [(0,2),(1,3)] and 
-- (NEq (g 0 1) (g 2 1)) reduced to [(0, 2)] and
-- (NEq (g 0 1) (g 1 0)) reduced to [(0, 1)]
-- we assume functions are represented as a particular kind of type constructor 
normalize1 :: [BCons] -> E [[(Int,Int)]]
normalize1 [] = Succ []
normalize1 ((NEq (Var i) (Var j)):xs) 
    | i == j    = Err ""                               -- x /= x  yields error
    | i < j     = report [(i,j)] (normalize1 xs)       -- put x /= y in order
    | otherwise = report [(j,i)] (normalize1 xs)
normalize1 ((NEq (Var i) (TyCon j ts)):xs)             -- we only keep [(i,j)] because
    | i < j     = report [(i,j)] (normalize1 xs)       -- x /= f y implies  x /= f z
    | otherwise = report [(j,i)] (normalize1 xs)
normalize1 ((NEq (TyCon i ts1) (TyCon j ts2)):xs)
    | i /= j    = normalize1 xs
    | (i==j) &&                                         -- type constructor can't be overloaded
      ts1==[]   = Err ""                                -- from here on we assume ts1,ts2 not nil
    | otherwise = let ns = foldl f [] (zip ts1 ts2)
                      f ls (t1,t2) = case (normalize1 [(NEq t1 t2)]) of   -- accumulate the normalized
                                      Succ l -> ((head l)++ls)            -- result of the arguments
                                      Err _  -> ls
                  in case ns of
                      [] -> Err ""                        -- if all arguments equal report error
                      _ -> report ((removeDups.sort) ns)  -- don't forget to keep lists sorted and
                                  (normalize1 xs)         -- remove duplicates
normalize1 ((NEq t1 t2):xs) = normalize1 ((NEq t2 t1):xs)


-- we omit further redundancies
-- [NEq 0 1, NEq (g 1 2) (g 0 3)] reduced to [NEq 0 1] and
-- [NEq 0 1, NEq 0 1] reduced to [NEq 0 1]
normalize2 :: [[(Int,Int)]] -> [[(Int,Int)]]
--normalize2 xs = (revreverse.normalize2'.revreverse.sort) xs
normalize2 xs = normalize2'' (sort xs) []                     -- more efficient

normalize2'' [] acc = acc
normalize2'' (x:xs) acc = let keep = foldl f True acc
                              f b y = b && (not (prefixmatch y x))
                          in if keep then normalize2'' xs (acc ++ [x])   -- keep order!
                             else normalize2'' xs acc

-}

prefixmatch [] _ = True
prefixmatch (i1:is1) (i2:is2)
   | i1 == i2  = prefixmatch is1 is2
   | otherwise = False

removeDups [] = []
removeDups (x:xs)
    | elem x xs = removeDups xs
    | otherwise = x:(removeDups xs)

--report r1 (Err _)   = Err ""
--report r1 (Succ r2) = Succ (r1:r2)



----------------------
-- Helper functions --
----------------------

-- assumption bc contains only equality constraints
buildUP :: [BCons] -> [(Type,Type)]
buildUP eq = foldl (\ eq -> \ (Eq t1 t2 _) -> ((t1,t2):eq)) [] eq

buildIEP :: [(Type,Type)] -> [BCons]
buildIEP [] = []
buildIEP ((t1,t2):ts) = (NEq t1 t2 noJust):(buildIEP ts)

buildEP :: [(Type,Type)] -> [BCons]
buildEP [] = []
buildEP ((t1,t2):ts) = (Eq t1 t2 noJust):(buildEP ts)


-- removes bcons of form: t = t  from list
remRedundantBCons :: [BCons] -> [BCons]
remRedundantBCons [] = []
remRedundantBCons (bc:bcs) | bcLeft bc == bcRight bc = rest
			   | otherwise = bc : rest
    where rest = remRedundantBCons bcs


--bconsLocs :: [BCons] -> [Loc]
bconsLocs bcs = let js = map bcJust bcs
		in  foldr addJust noJust js
