--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module Chr(Index, PredSym, CHR(..),
	      matchCons, subsetCons, 
	      simpName, simpNameStr, chSimpName, simpUCons, chrRhs, chrLhs,
	      isProp, isSimp, isMSimp,chrAddCons,
	      chrNewRhs,  groupCHRs, rhsToCons, consToRhs, newSimpUCons,
	      simpToMSimp, uconsPSym, 
	      stdInfo, noInfo, joinInfo, defInfo, annInfo, taddInfo, mdefInfo,
	      odefInfo, oannInfo, oinfInfo, o0Info, ruleInfo, chrType,
	      iannInfo, oruleInfo, chrLoc, chrTypeId, isDefCHR, isMDefCHR,
	      isStructCHR, isAnnCHR, fitsAnnCHR, isValueMDef, CHRInfo,
	      presentableCHR, isValueDef, isOverloadDef, isOverloadInf,
	      isValJoinCHR, isOverloadJoinCHR, chrId, isOverloadRule, isValCHR, 
	      isValueAnn, isValueTAdd, isNonOverAnnCHR, stripGlobalUCons,
	      isInfAnnCHR, isOverloadCHR,
	      omdefInfo, ojoinInfo, isOverloadMDef, CHRInfoFun, isJoinCHR,
	      emptyRhs, isCheckableAnnCHR,
	      chrInfo, chrSrcPos, newCHRSrcPos,
	      
	      chrLambdaVec, chrStvVec,
	      infAnnToAnnCHR, infAnnUCons )
	      
where
 
import Core
import UCons	(Index)	    -- not in Core: no other module needs it
import Ann_InputAST hiding (Index)	-- for ShowAnn (move it out!)
import Pretty
import Terminal
import Misc

import Justify
import List
import Char
import IOExts

-------------------------------
-- Type and data definitions --
------------------------------- 



-- CHR rules:
-- Prop rules are indexed to prevent infinite application and
-- head atoms are sorted to handle commutativity (not done yet!)


data CHR = Simp  CHRInfo UCons [BCons] ([UCons],[BCons]) |
	   MSimp CHRInfo [UCons] [BCons] ([UCons],[BCons]) |
           Prop  CHRInfo Index [UCons] [BCons] ([UCons],[BCons])    deriving Eq

-- information about the source and the purpose of a CHR
data CHRInfo = CHRInfo CHRType Loc SrcPos	deriving (Show, Eq)
data CHRType = MiscCHR | Rule | Standard
	     | ValueDef Int Id | ValueMDef Int Id 
	     | ValueAnn Int Id  [UCons]	-- ::  (Global UCons!)
	     | ValueTAdd Int Id [UCons]	-- :::
	     | ValueInfAnn Int Id [UCons] UCons	-- an annotation CHR generated
						-- from the inference CHR
	     | ValueJoin Int Id 
	     | OverloadDef Int Id | OverloadAnn Int Id | OverloadInf Id 
	     | OverloadMDef Int Id | OverloadJoin Int Id
	     | OverloadRule Id | Overload0 Id    deriving (Show, Eq)

type CHRInfoFun = Id -> Loc -> CHRInfo

mkCHRInfo x y = CHRInfo x y noSrcPos

noInfo  = mkCHRInfo MiscCHR noLoc
stdInfo = mkCHRInfo Standard noLoc
defInfo  n id l = mkCHRInfo (ValueDef n id) l
annInfo  n id l ucs = mkCHRInfo (ValueAnn n id ucs) l
taddInfo n id l ucs = mkCHRInfo (ValueTAdd n id ucs) l
iannInfo n id l ucs uc = mkCHRInfo (ValueInfAnn n id ucs uc) l
mdefInfo n id l = mkCHRInfo (ValueMDef n id) l
joinInfo n id l = mkCHRInfo (ValueJoin n id) l
odefInfo n id l = mkCHRInfo (OverloadDef n id) l
oannInfo n id l = mkCHRInfo (OverloadAnn n id) l
ojoinInfo n id l = mkCHRInfo (OverloadJoin n id) l
omdefInfo n id l = mkCHRInfo (OverloadMDef n id) l
oinfInfo id l = mkCHRInfo (OverloadInf id) l
oruleInfo id l = mkCHRInfo (OverloadRule id) l
o0Info id l = mkCHRInfo (Overload0 id) l
ruleInfo l = mkCHRInfo Rule l


isRule n = case n of Rule -> True; _ -> False
isValueDef n = case n of (ValueDef _ _) -> True; _ -> False
isValueAnn n = case n of (ValueAnn _ _ _) -> True; _ -> False
isValueInfAnn n = case n of (ValueInfAnn _ _ _ _) -> True; _ -> False
isValueTAdd n = case n of (ValueTAdd _ _ _) -> True; _ -> False
isValueMDef n = case n of (ValueMDef _ _) -> True; _ -> False
isValueJoin n = case n of (ValueJoin _ _) -> True; _ -> False
isOverloadDef n = case n of (OverloadDef _ _) -> True; _ -> False
isOverloadAnn n = case n of (OverloadAnn _ _) -> True; _ -> False
isOverloadInf n = case n of (OverloadInf _) -> True; _ -> False
isOverloadMDef n = case n of (OverloadMDef _ _) -> True; _ -> False
isOverloadJoin n = case n of (OverloadJoin _ _) -> True; _ -> False
isOverloadRule n = case n of (OverloadRule _) -> True; _ -> False


-- true if a chr from a rule dec or overload annotation
isStructCHR :: CHR -> Bool
isStructCHR chr = let n = chrType chr
		  in  isRule n || isOverloadRule n

isValJoinCHR :: CHR -> Bool
isValJoinCHR chr = isValueJoin (chrType chr)

isOverloadJoinCHR :: CHR -> Bool
isOverloadJoinCHR chr = isOverloadJoin (chrType chr)

isJoinCHR :: CHR -> Bool
isJoinCHR chr = isValJoinCHR chr || isOverloadJoinCHR chr

isMDefCHR :: CHR -> Bool
isMDefCHR chr = let n = chrType chr
	        in  isValueMDef n || isOverloadMDef n

isDefCHR :: CHR -> Bool
isDefCHR chr = let n = chrType chr
	       in  isValueDef n || isOverloadDef n

-- true if chr is from an overload or value annotation
isAnnCHR :: CHR -> Bool
isAnnCHR chr = let n = chrType chr
	       in  isValueAnn n || isValueTAdd n || isOverloadAnn n

isInfAnnCHR chr = isValueInfAnn (chrType chr)

isNonOverAnnCHR :: CHR -> Bool
isNonOverAnnCHR chr = let n = chrType chr
		      in  isValueAnn n || isValueTAdd n

isCheckableAnnCHR :: CHR -> Bool
isCheckableAnnCHR chr = let n = chrType chr
			in  isValueAnn n || isOverloadAnn n

isValCHR :: CHR -> Bool
isValCHR chr = 
    let n = chrType chr
    in  isValueDef n || isValueAnn n || isValueMDef n || isValueJoin n

isOverloadCHR :: CHR -> Bool
isOverloadCHR chr = 
    let n = chrType chr
    in  isOverloadDef n  || isOverloadAnn n  || isOverloadInf n ||
	isOverloadMDef n || isOverloadJoin n || isOverloadRule n

infAnnToAnnCHR :: CHR -> CHR
infAnnToAnnCHR (Simp (CHRInfo (ValueInfAnn n id ucs _) l sp) luc lbcs rhs) = 
	      Simp (CHRInfo (ValueAnn n id ucs) l sp) luc lbcs rhs
infAnnToAnnCHR chr = chr

infAnnUCons :: CHR -> UCons
infAnnUCons (Simp (CHRInfo (ValueInfAnn _ _ _ uc) _ _) _ _ _) = uc
infAnnUCons _ = error "infAnnUCons"

-- second chr is an ann chr, first an arbitrary chr
-- True if the first chr is the mdef of the definition which the ann applies to
fitsAnnCHR :: CHR -> CHR -> Bool
fitsAnnCHR chr achr = 
    let c = chrType chr
	a = chrType achr
    in  (isValueJoin c || isOverloadJoin c) && 
	(chrTypeId c == chrTypeId a && chrTypeNum c == chrTypeNum a)


-- if passed an ann. CHR, removes the global constraints from it
stripGlobalUCons :: CHR -> CHR
stripGlobalUCons chr = case chr of
    Simp (CHRInfo (ValueAnn n id gucs) l sp) luc lbcs (rucs,rbcs) -> 
	Simp (CHRInfo (ValueAnn n id gucs) l sp) luc lbcs (rucs `rem` gucs,rbcs)
    Simp (CHRInfo (ValueTAdd n id gucs) l sp) luc lbcs (rucs,rbcs) ->
	Simp (CHRInfo (ValueTAdd n id gucs) l sp) luc lbcs (rucs`rem` gucs,rbcs)
    _ -> chr

    where rem ucs gucs = [ uc | uc <- ucs, uc `notElem` gucs ]

chrId chr = chrTypeId (chrType chr)

chrTypeId (ValueDef _ id)    = id
chrTypeId (ValueAnn _ id _)  = id
chrTypeId (ValueMDef _ id)   = id
chrTypeId (ValueJoin _ id)   = id
chrTypeId (OverloadDef _ id)  = id
chrTypeId (OverloadAnn _ id)  = id
chrTypeId (OverloadMDef _ id) = id
chrTypeId (OverloadInf id) = id
chrTypeId (Overload0 id)   = id
chrTypeId (OverloadJoin _ id) = id
chrTypeId _		   = ""

chrTypeNum (OverloadDef n _)  = n
chrTypeNum (OverloadAnn n _)  = n
chrTypeNum (OverloadMDef n _) = n
chrTypeNum (ValueDef n _)   = n
chrTypeNum (ValueMDef n _)  = n
chrTypeNum (ValueAnn n _ _) = n
chrTypeNum (ValueTAdd n _ _)= n
chrTypeNum (ValueJoin n _)  = n
chrTypeNum (OverloadJoin n _) = n
chrTypeNum _		      = noLoc


newCHRTypeNum :: Int -> CHR -> CHR
newCHRTypeNum n chr = 
    let info = case chrInfo chr of
	 CHRInfo t l sp -> 
		        let t' = case t of
			     ValueDef _ id  -> ValueDef n id
			     ValueMDef _ id -> ValueMDef n id
			     ValueAnn _ id ucs -> ValueAnn n id ucs
			     ValueTAdd _ id ucs -> ValueTAdd n id ucs
			     ValueJoin _ id -> ValueJoin n id
			     OverloadDef _ id -> OverloadDef n id
			     OverloadAnn _ id -> OverloadAnn n id
			     OverloadMDef _ id -> OverloadMDef n id
			     OverloadJoin _ id -> OverloadJoin n id
			     _ -> t
			in  CHRInfo t' l sp
    in  newCHRInfo info chr

chrStvVec = error "chrStvVec"
chrLambdaVec = error "chrLambdaVec" 

----------------------------------------

isProp (Prop _ _ _ _ _) = True
isProp _	      = False
isSimp (Simp _ _ _ _)   = True
isSimp _	      = False
isMSimp (MSimp _ _ _ _) = True
isMSimp _	      = False

-- partitions a list of CHRs into simps, msimps and props
groupCHRs :: [CHR] -> ([CHR], [CHR], [CHR])
groupCHRs [] = ([], [], [])
groupCHRs (c:cs) = let (s, m, p) = groupCHRs cs
		   in  case c of
			r@(Simp _ _ _ _)   -> (r:s, m, p)
			r@(MSimp _ _ _ _)  -> (s, r:m, p)
			r@(Prop _ _ _ _ _) -> (s, m, r:p)
	
simpToMSimp (Simp n uc bcs rhs) = MSimp n [uc] bcs rhs
simpToMSimp chr	= trace "simpToMSimp: arg not a simp!" chr

instance PPrint CHR where

  pprint (Simp n uc x rhs)    = 
		 let duc  = show uc
		     drhs = show (rhsToCons rhs)
		     dn = show n
		     dx = show x
		 in  return (text (duc ++ " <==> " ++ drhs)) --  ++ "|" ++ dn))
		 
  pprint (MSimp n ucs _ rhs)  = 
		 let ducs = uncommas (map show ucs)
		     drhs = show (rhsToCons rhs)
		 in  return (text (ducs ++ " <==> " ++ drhs))

  pprint (Prop n _ ucs _ rhs) = 
		 let ducs = uncommas (map show ucs)
		     drhs = show (rhsToCons rhs)
		 in  return (text (ducs ++ " ==> " ++ drhs))

instance Show CHR where show x = pprintPlain x

{-
instance Show CHR where
  show (Simp n ucon _ ubcons)    = {- "(" ++ show n ++ ") " ++ -} show ucon ++ " <==> " ++ showUBCons ubcons
  show (MSimp n ucons _ ubcons)  = {- "(" ++ show n ++ ") " ++ -} commaList ucons ++ " <==> " ++
				   showUBCons ubcons
  show (Prop n _ ucons _ ubcons) = {- "(" ++ show n ++ ") " ++ -} commaList ucons ++ " ==> " ++ 
				   showUBCons ubcons

instance ShowAnn CHR where
  showAnn (Simp n luc _ (rucs, rbcs)) = 
					"(" ++ show n ++ ") " ++
					showAnn luc  ++ " <==> " ++
				        annCommaList rucs ++ 
				        annCommaList rbcs
  showAnn (MSimp n lucs _ (rucs, rbcs)) = "(" ++ show n ++ ") " ++
					  showAnn lucs ++ " <==> " ++
					  annCommaList rucs ++
					  annCommaList rbcs 
  showAnn (Prop n _ lucs lbcs (rucs, rbcs)) = 
					   "(" ++ show n ++ ") " ++
					   annCommaList lucs ++
					   annCommaList lbcs ++ " ==> " ++
					   annCommaList rucs ++
					   annCommaList rbcs
-}
showUBCons ([],[])	= "True"
showUBCons (us,[])	= commaList us
showUBCons ([],bs)	= commaList bs
showUBCons (us,bs)	= commaList us ++ ", " ++ commaList bs

-- CHR store(s)

type Goal  = ([UCons],[BCons])
type Token = [Index]                     -- index of rule number and atoms    
                                         -- assumption: sorted in non-decreasing order 
type TokenSet =[Token] 
-- ucons must be sorted to handle commutativity
-- this is an invariant which needs to be preserved
type Store = (Goal,[UCons], [BCons],TokenSet, [TVar])

storeUCons (_, ucs, _, _, _) = ucs
storeBCons (_, _, bcs, _, _) = bcs

---------------------------
-- Instance declarations --
---------------------------

-- Note: methods don't work on indices!
-- see rnIdx below!

instance TC CHR where
   apply s (Simp n uc bcs1 (ucs2,bcs2)) = 
       Simp n (apply s uc) (apply s bcs1) (apply s ucs2,apply s bcs2)
   apply s (MSimp n ucs bcs1 (ucs2,bcs2)) =
          MSimp n (apply s ucs) (apply s bcs1) (apply s ucs2,apply s bcs2)    
   apply s (Prop n i ucs1 bcs1 (ucs2,bcs2)) =
       Prop n i (apply s ucs1) (apply s bcs1) (apply s ucs2,apply s bcs2)    

   fv (Simp _ uc bcs1 (ucs2,bcs2)) = (fv uc)++(fv bcs1)++(fv ucs2)++(fv bcs2)
   fv (MSimp _ ucs bcs1 (ucs2,bcs2)) = (fv ucs)++(fv bcs1)++(fv ucs2)++(fv bcs2)
   fv (Prop _ i ucs1 bcs1 (ucs2,bcs2)) = (fv (ucs1++ucs2)) ++ (fv (bcs1++bcs2))
   
   occurs v (Simp _ uc bcs1 (ucs2,bcs2)) = (occurs v (bcs1++bcs2)) || (occurs v (uc:ucs2))
   occurs v (MSimp _ ucs1 bcs1 (ucs2,bcs2)) = (occurs v (bcs1++bcs2)) || (occurs v (ucs1++ucs2))
   occurs v (Prop _ i ucs1 bcs1 (ucs2,bcs2)) = (occurs v (bcs1++bcs2)) || (occurs v (ucs1++ucs2))

   rename (Simp n uc bcs1 (ucs2,bcs2)) = 
          do s <- renameSubst ((fv (uc:ucs2))++(fv (bcs1++bcs2)))
             return (Simp n (apply s uc) (apply s bcs1) (apply s ucs2,apply s bcs2))
   rename (MSimp n ucs1 bcs1 (ucs2,bcs2)) = 
          do s <- renameSubst ((fv (ucs1++ucs2))++(fv (bcs1++bcs2)))
             return (MSimp n (apply s ucs1) (apply s bcs1) (apply s ucs2,apply s bcs2))	     
   rename (Prop n i ucs1 bcs1 (ucs2,bcs2)) =
          do s <- renameSubst ((fv (ucs1++ucs2)) ++ (fv (bcs1++bcs2)))
             return (Prop n i (apply s ucs1) (apply s bcs1) (apply s ucs2,apply s bcs2))
	     
   format (Simp _ uc bcs1 (ucs2,bcs2)) =
           let text1 = if bcs1 == [] then []
                       else (format bcs1) ++ " | "
               text2 = if ucs2 == [] then []
                       else (format ucs2) ++ ","
           in (format uc) ++ "<==> " ++ text1 ++ text2 ++ (format bcs2)	   
   format (MSimp _ ucs1 bcs1 (ucs2,bcs2)) =
           let text1 = if bcs1 == [] then []
                       else (format bcs1) ++ " | "
               text2 = if ucs2 == [] then []
                       else (format ucs2) ++ ","
           in (format ucs1) ++ "<==> " ++ text1 ++ text2 ++ (format bcs2)
   format (Prop _ i ucs1 bcs1 (ucs2,bcs2)) =
           let text1 = if bcs1 == [] then []
                       else (format bcs1) ++ " | "
               text2 = if ucs2 == [] then []
                       else (format ucs2) ++ ","
           in "Idx" ++ (show i) ++ ": " ++ (format ucs1) ++ "==> " ++ 
              text1 ++ text2 ++ (format bcs2)         



----------------------
-- Helper functions --
----------------------

simpName (Simp _ (UC _ nm _ _) _ _)  = nm
simpName s = error $ "TD_Chr.simpName: " ++ show s

simpNameStr s = let Name (str,_) = simpName s in str

chSimpName nm (Simp n (UC ix _ ts x) rbcs rhs)
                    = Simp n (UC ix nm ts x) rbcs rhs
chSimpName _  _     = error "TD_Chr.chSimpName"

simpUCons (Simp _ uc _ _) = uc
simpUCons _ = error "TD_Chr.simpUCons"

newSimpUCons (Simp n _ bc rhs) uc = Simp n uc bc rhs
newSimpUCons _ _ = error "newSimpUCons"

chrNewRhs rhs (Simp n u b _)   = Simp n u b rhs
chrNewRhs rhs (MSimp n u b _)  = MSimp n u b rhs
chrNewRhs rhs (Prop n i u b _) = Prop n i u b rhs

chrRhs (Simp _ _ _ rhs)   = rhs
chrRhs (MSimp _ _ _ rhs)  = rhs
chrRhs (Prop _ _ _ _ rhs) = rhs

consToRhs c = (consToUs c, consToBs c)
rhsToCons (ucs, bcs) = usToCons ucs `cand` bsToCons bcs
emptyRhs = ([], [])

chrLhs (Simp _ u b _)    = [u]
chrLhs (Prop _ _ us b _) = us 
chrLhs (MSimp _ us b _)  = us

chrInfo (Simp n _ _ _)   = n
chrInfo (MSimp n _ _ _)  = n
chrInfo (Prop n _ _ _ _) = n

newCHRInfo n (Simp _ a b c) = Simp n a b c
newCHRInfo n (MSimp _ a b c)  = MSimp n a b c
newCHRInfo n (Prop _ a b c d) = Prop n a b c d


chrLoc  chr = case chrInfo chr of (CHRInfo _ l _) -> l
chrType chr = case chrInfo chr of (CHRInfo t _ _) -> t
chrSrcPos chr = case chrInfo chr of (CHRInfo _ _ sp) -> sp

newCHRSrcPos sp chr = let CHRInfo t l _ = chrInfo chr
			  info' = CHRInfo t l sp
		      in  newCHRInfo info' chr

instance Justifiable CHR where
    unJust (Simp _ uc _ _)   = unJust uc
    unJust (MSimp _ uc _ _)  = unJust uc
    unJust (Prop _ _ uc _ _) = unJust uc

    reJust j (Simp a uc b rhs)   = Simp a (reJust j uc) b (reJust j rhs)
--    reJust (MSimp _ uc _ rhs)  = reJust uc
--    reJust (Prop _ _ uc _ rhs) = reJust uc
    

-- takes a rule and a constraint, and adds the constraint to the rule's body
chrAddCons (Simp n luc lbcs (rucs, rbcs)) cons = 
    let bcs = consToBs cons
	ucs = consToUs cons
    in  Simp n luc lbcs (ucs ++ rucs, bcs ++ rbcs)
chrAddCons (MSimp n lucs lbcs (rucs, rbcs)) cons =
    let bcs = consToBs cons
	ucs = consToUs cons
    in  MSimp n lucs lbcs (ucs ++ rucs, bcs ++ rbcs)
chrAddCons (Prop n ix lucs lbcs (rucs, rbcs)) cons = 
    let bcs = consToBs cons
	ucs = consToUs cons
    in  Prop n ix lucs lbcs (ucs ++ rucs, bcs ++ rbcs)


-- renames the variables in the CHR so that they have simple, readable names
presentableCHR :: CHR -> CHR
presentableCHR chr = let vs  = nub (fv chr)
			 phi = simpTVarSubst vs
		     in  apply phi chr
