--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module Compiler (compile, compileChecks)
where

import Ann_InputAST (Dec(..), items, strip)
import TD_General (instToOverDec, instDecs)
import EvidenceTrans
import TD_Monad
import OptFlags (Flag(..), outputFlag, verbose)

import IO
import Monad
import Terminal


--------------------------------------------------------------------------------

-- convert instance declarations into overload declarations before invoking the
-- evidence translation

compile :: [Flag] -> TD ()
compile fs = do
    (h,f) <- outHandle fs
    fixProg
    verb (puts "> evidence translation")
    evidenceTranslation
    hd <- header
    p  <- getProg
    let output = hd ++ show (strip p)
    doIO (hPutStr h output)
    
    where fixProg = do
	    Prog adecs <- getProg
	    setProg (Prog (map instToOverDec adecs))
    
	  outHandle fs = do
	    case outputFlag fs of
		Nothing	-> do h <- doIO (openFile stdFile WriteMode)
			      return (h, stdFile)
		Just f  -> do h <- doIO (openFile f WriteMode) 
			      return (h, f)
	  stdFile = "out.hs"

	  -- generates a header for the translated output
	  header :: TD String
	  header = do hns <- getHaskellConNames
		      let cs = map (\n -> "class " ++ show n ++ " a") hns
		      -- return (unlines cs ++ "\n")
		      return ""
	  
	  verb = when (verbose fs)

----------------------------------------

-- checks which should be performed after all other static analysis, but just
-- before compilation

compileChecks :: [Flag] -> TD ()
compileChecks fs = do
    Prog adecs <- getProg
    let ids  = instDecs adecs
	ids' = [ d | d@(InstDec _ ds) <- items ids, null ds ]
    when (not (null ids')) $
	puts ("Warning: Following instance " ++ plural ids' "declaration" ++
	      " lack definition:\n" ++ unlines (map (('\t':).show) ids'))

