--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module ConsReason where

import Constraint
import Justify
import Types
import List
import IOExts
import Misc
import Ann_InputAST (Loc, unkIx)

import Name


-- predicate on sets of some type
type Pred a = a -> Bool

-- predicate on sets of eq. constraints
type BCPred = Pred [BCons]

-- predicate for incremental tests
-- extends subst with bcons if successful, else fails
type BCPredInc = (IncState, BCons) -> (IncState, Bool)

type IncState = (E Subst, [BCons])

newIncState :: IncState
newIncState = (Succ [], [])

updateIncState :: IncState -> BCons -> IncState
updateIncState (es, bcs) bc@(Eq t1 t2 _) =
    let es' = unify [] (unSucc es) [(t1, t2)]
    in  (es', bc:bcs)
    
-- grouped BCons
type GBCons = ([Loc], [BCons])

--------------------------------------------------------------------------------

{-
-- True if bc1 entails bc2
entails :: [BCons] -> [BCons] -> Bool
entails bc1 bc2@[Eq (Var _) (Var x) _]
-- hacky, check why this doesn't follow from entails' (it should)
    | x `notElem` (fv bc1) = trace "%%%%%%%%%%%%%%%%%%" True   
    | otherwise		   = trace "$$$$$$$$$$$$$$$$$$" $ entails' bc1 bc2
entails bc1 bc2 = trace "$$$$$$$$$$$$$$$$$$$" $ entails' bc1 bc2
entails' bc1 bc2 = 
    let s1 = buildMGU bc1
	s2 = buildMGU bc2
    in  if isErr s1 || isErr s2 || unsatBCs (apply (unSucc s1) bc2)
	   then False
	   else let s1'  = unSucc s1
		    s2'  = unSucc s2
		    vs1  = map fst s1'
		    s2'' = [ m | m@(v, _) <- s2', v `velem` vs1 ]
		    vs   = map fst s2''
		    s1'' = [ m | m@(v, _) <- s1', v `velem` vs ]
		    ps   = pair s1'' s2''
		    free = and [ v `notElem` (fv bc1) | v <- fv bc2 ]
		    ent  = -- free || (not (null ps) && all asGround ps)
			    (trace "A" free) || (trace "B" (null ps)) || 
			    (trace "C" all asGround ps)
		    msg  =  "\ns1': " ++ show s1' ++ 
			    "\ns2': " ++ show s2' ++
			    "\ns1'': " ++ show s1'' ++
			    "\ns2'': " ++ show s2'' ++
			    "\nps: " ++ show ps
		    in trace msg 
			ent
    where 
	veq (v1, _) (v2, _) = v1 == v2
	velem v vs = any (v`veq`) vs
	notVElem v vs = not (velem v vs)
	vlookup v []		= Nothing
	vlookup v ((v',t):vts)
	    | v `veq` v' = Just t
	    | otherwise	 = vlookup v vts 
	    
    
	-- assuming sat(bc1 ++ bc2)

        asGround (t1, t2) = fst (asGround' [] (t1, t2))
            where 
                asGround' vs (Var a, Var b) = 
                    case lookup b vs of
                         Nothing    -> (True, (b, a):vs)
                         Just c     -> (c == a, vs)
                         
                asGround' vs (TApp t1 t2, TApp s1 s2) = 
                    let (b1, vs1) = asGround' vs  (t1,s1)
                        (b2, vs2) = asGround' vs1 (t2,s2)
                    in  (b1 && b2, vs2)
                asGround' vs (Var _, _) = (False, vs)
                asGround' vs (_, _)     = (True, vs)


{-

	asGround (Var _, Var _) = True
	asGround (TApp t1 t2, TApp s1 s2) = asGround (t1,s1) && 
					    asGround (t2,s2)
	asGround (Var _, _) = False 
	asGround (_, _)	    = True
-}
	
	buildMGU bcs = 
	    let vs = fv bcs
	    in  case unifyWithIds (fv bcs) idSubst (buildUP bcs) of
		    s@(Succ mgu)  -> s -- Succ (extend vs mgu)
		    e@(Err _) -> e
	--extend vs mgu = mgu ++ [ (v,Var v) | v <- vs, v `notVElem` map fst mgu ]
	--extend = flip const  
	pair [] _ = []
	pair ((v,t):vts) s2 = let Just t' = vlookup v s2
			      in  (t, t') : pair vts s2

-}

--entails :: [BCons] -> [BCons] -> Bool
--entails bs1 bs2 = entails' (fv bs1) bs1 bs2
--entails bs1 bs2 = entails' (nub (fv bs1 ++ fv bs2)) bs1 bs2



-- WARNING: useful only for explain command (because of assumption that vars on
-- the lhs of eqs being entailed are important)
entails :: [BCons] -> [BCons] -> Bool
entails bs1 bs2 = 
    let vs1 = fv bs1 
	ent = entails' vs1 bs1 bs2
	lvs2 = fv (map bcLeft bs2)
	gotVs = all (`elem`vs1) lvs2
    in  ent && gotVs


-- takes a list of constraints, and a list of equalities (represeted by two
-- parallel lists) and returns true if there exists phi s.t.:  
--  bcs |= phi ls <-> rs
entails2l :: [BCons] -> [Type] -> [Type] -> Bool
entails2l bcs ls rs =
    let Succ s = unify [] idSubst (buildUP bcs)
	rs' = map (apply s) rs
	a = matchTypes ls rs' 
	a' = trace ("entails2l  ls: " ++ show ls ++ "\trs': " ++ show rs' ++
		    "\tmt: " ++ show a) a
    in  a'


entails2 :: [BCons] -> [(Type, Type)] -> Bool
entails2 bcs pts = 
    let (ls, rs) = unzip pts
	Succ s = unify [] idSubst (buildUP bcs)
	ls' = map (apply s) ls
	pts' = zip ls' rs
	res  = and [ matchTypes [l] [r] | (l, r) <- pts' ]
    in  res



-- True if some subset of c1 entails c2, c1 doesn't have to be satisfiable
-- if it's not, have to go through all *maximal* satisfiable subsets until we
-- either get to one which does entail c2, or exhaust all possibilities (in
-- which c1 doesn't entail c2)
-- If successful, returns the subset which does entail c2, else Nothing
subsetEntails :: [BCons] -> [BCons] -> Maybe [BCons] 
subsetEntails c1 c2 =  let ss = satisfiableSubsets c1
		           es = [ c | c <- ss, c `entails` c2 ]
		       in  case es of
			    (x:_)   -> Just x
			    _	    -> Nothing
    

-- given an unsatisfiable set of constraints, returns all satisfiable subsets
-- WARNING: currently only works if there are constraints common to all
--	    unsatisfiable subsets!
satisfiableSubsets :: [BCons] -> [[BCons]]
satisfiableSubsets bcs
    | satBCs bcs    = [bcs]
    | otherwise	    = [ bs | (_, bs) <- allLeft bcs, satBCs bs ]
    

--entails' :: [BCons] -> [BCons] -> Bool
entails' vs bs1 bs2 = 
    let mphi2 = unify [] idSubst (buildUP bs2)
    in  if isErr mphi2
	   then False
	   else let phi2 = unSucc mphi2
		    bs1' = apply phi2 bs1

		    mphi1  = unify [] idSubst (buildUP bs1)
		    mphi1' = unify [] idSubst (buildUP (bs2 ++ bs1)) -- bs1')

		in  if isErr mphi1 || isErr mphi1'
		       then False
		       else let phi1  = unSucc mphi1
				-- phi1' = unSucc mphi1'
				phi1' = phi1 ++ phi2
			        res = and [ ag | v <- vs, 
						 let t  = Var v,
						 let v1 = apply phi1 t,
						 let v2 = apply phi2 t,
						 let ag = asGround (v1, v2) ]
			    in res 
			   
    where
--	asGround (t1, t2) = matchTypes [t1] [t2]
    
        asGround (t1, t2) = fst (asGround' [] (t1, t2))
            where 
                asGround' vs (Var a, Var b) = 
                    case lookup b vs of
                         Nothing    -> (True, (b, a):vs)
                         Just c     -> (c == a, vs)
                         
                asGround' vs (TApp t1 t2, TApp s1 s2) = 
                    let (b1, vs1) = asGround' vs  (t1,s1)
                        (b2, vs2) = asGround' vs1 (t2,s2)
                    in  (b1 && b2, vs2)
                asGround' vs (Var _, _) = (False, vs)
                asGround' vs (_, _)     = (True, vs)


-- used for rule matching / justification propagation
entailType :: [BCons] -> BCons -> Bool
entailType bcs bc = bcs `entails` [bc]
	


-- used for type explanations
entailTypeExp :: [BCons] -> BCons -> Bool
--entailTypeExp bcs bc@(Eq (Var v) t _) = 
--    (bcs `entails` [bc]) && (v `elem` (fv bcs))

entailTypeExp bcs bc = any (`elem`(fv bcs)) (fv bc) && bcs `entails` [bc]



-- tests whether the given constraint is satisfiable or not
sat, unSat :: Constraint -> Bool
sat   c = not (unSat c)
unsat   = unSat
unSat c = let bcs = consToBs c
	  in  unSatBCs bcs


satBCs, unSatBCs :: BCPred
satBCs   bcs = not (unSatBCs bcs)
unsatBCs     = unSatBCs
unSatBCs bcs = let eqp = buildUP bcs
		   mgu = unify [] idSubst eqp
	       in  isErr mgu
	       
satBCsInc, unsatBCsInc :: BCPredInc
satBCsInc isb = let (is, b) = unsatBCsInc isb
		in  (is, not b)
unsatBCsInc (is, bc) = let is'@(es, _) = updateIncState is bc
		       in  (is', isErr es)


--sat' :: Constraint -> TD Bool
--sat' solver chrs n cons = let solve = solver chrs cons
--		          in  isSucc (unTP solve n)
sat' solver chrs n cons = unTP (solver chrs cons) n
		          
unSat' s ch n c = let st  = sat' s ch n c
		      res = not (isSucc st)
		  in  trace ("constraint: " ++ show c ++ "\n" ++
			     "\tstore:" ++ show st ++ "\n" ++
			     "\tunSat?" ++ show res ++ "\n") res


bcHasJust l bc = l `inJust` (bcJust bc)
ucHasJust l uc = l `inJust` (ucJust uc)


singGroupCons :: Constraint -> [(Loc, Constraint)]
singGroupCons c = singGroupConsExcept c []

-- takes a constraint, and breaks it into a list of constraints, where the
-- constraint in each element are justified by a common location
-- we generate such a list for each location we find in the Constraint, except 
-- for those passed to groupConsExcept
singGroupConsExcept :: Constraint -> [Loc] -> [(Loc, Constraint)]
singGroupConsExcept c ls = 
		       let ucs = consToUs c
			   bcs = consToBs c
			   lus = concatMap (unJust.ucJust) ucs
			   lbs = concatMap (unJust.bcJust) bcs
			   ls'  = filter (`notElem` ls) (nub (lus ++ lbs))
			   g_ucs  = [ filter (ucHasJust l) ucs | l <- ls' ] 
			   g_bcs  = [ filter (bcHasJust l) bcs | l <- ls' ]
			   groups = [ (usToCons ucs) `cand` (bsToCons bcs) | 
					    (ucs, bcs) <- zip g_ucs g_bcs ]
		       in zip ls' groups

-- takes a constraint and a list of locations, and returns a sub-constraint 
-- which is justified by at least one of those locations
consFromLoc :: Constraint -> [Loc] -> Constraint
consFromLoc c ls = let ucs = consToUs c
		       bcs = consToBs c
		       ucs' = filter (\uc -> any (`ucHasJust` uc) ls) ucs
		       bcs' = filter (\bc -> any (`bcHasJust` bc) ls) bcs
		   in  bsToCons bcs' `cand` usToCons ucs'

-- removes constraints justified by the location
remConsLoc c l = let ucs = consToUs c
                     bcs = consToBs c
                     ucs' = filter (not . (ucHasJust l)) ucs
                     bcs' = filter (not . (bcHasJust l)) bcs
                 in  bsToCons bcs' `cand` usToCons ucs'

----------------------------------------

-- takes a constraint, and breaks it into a list of constraints, where the
-- constraint in each element are justified by a common location
-- we generate such a list for each set of locations we find in the Constraint
{-
groupCons :: Constraint -> [([Loc], Constraint)]
groupCons c = let ucs = consToUs c
		  bcs = consToBs c
		  jus = map ucJust ucs
		  jbs = map bcJust bcs
		  ls'  = nub (map unJust (jus ++ jbs))
		  ls   = filter (not . null) ls'    -- FIXME: HACK!!!
		  grps = [ (l, bsToCons (getBs l bcs) `cand` 
			       usToCons (getUs l ucs)) | l <- ls ]
	      in  grps
    where getBs l bs = filter (\b -> unJust (bcJust b) == l) bs
	  getUs l us = filter (\u -> unJust (ucJust u) == l) us
-}

groupBCons :: [BCons] -> [GBCons]
groupBCons bcs = let js   = map bcJust bcs
		     ls   = nub (map unJust js)
		     grps = [ (l, getBs l bcs) | l <- ls ]
		 in  grps
    where getBs l bs = filter (\b -> unJust (bcJust b) == l) bs 

----------------------------------------
-- given two satisfiable constraints, c1, c2... returns the minimal subset of
-- c1 that entails c2
entailSubset :: Constraint -> BCons -> [[Loc]]
entailSubset c bc = 
    let bcss = minImplSubsets (consToBs c) bc
	bcss2 = elimUnsats bcss
	bcss2' = trace ("\nconstraints: " ++ show bcss ++
			"\nentail: " ++ show bc) bcss2
    in  [ js | bcs <- bcss2, let bcs' = map bcJust bcs,
			     let js = unJust (foldr addJust noJust bcs')]

    where elimUnsats [] = []
	  elimUnsats (bcs:bcss)
	    | unsatBCs (bc:bcs) = elimUnsats bcss
	    | otherwise		= bcs : elimUnsats bcss


-- used for explanations
entailSubsetExp :: Constraint -> BCons -> [[Loc]]
entailSubsetExp = entailSubset


-- use for explaining subsumption error
-- (none of the variables appearing in the bc are fresh!)
entailSubsetSub :: Constraint -> BCons -> [BCons] -- [[Loc]]
entailSubsetSub c bc = 
	let p bcs = entailBCons bcs (fv bc) [bc]
	    bcs = consToBs c
	in  oneMinSubsetBCons p bcs

--------------------------------------------------------------------------------
-- miscellaneous.

-- may be that two different subsets are found, but the justifications
-- for one are a superset of the justs for the other... elimByJust 
-- gets rid of such sets
elimByJust xss = 
    let lss  = map unJust xss
	sups = [ any (ys`superset`) yss | (ys, yss) <- allLeft lss ]
   in  [ zs | (zs, b) <- zip xss sups, not b ]

--------------------------------------------------------------------------------

-- for an inconsistent bcs, returns all the constraints which appear in *all*
-- minimal unsatisfiable subsets
-- flag is True if this set is unsatisfiable (which would make it the only
-- minimal unsat. subset)
commonUnsatCons :: [BCons] -> ([BCons], Bool)
commonUnsatCons bcs = 
    let com = [ b | (b, bs) <- allLeft bcs, satBCs bs ]
    in  (com, unsatBCs com)

-- as above, except the set of initial constraints is given in two parts
-- (the complete store is given by bcs1 ++ bcs2)
-- when looking for common constraints, we ONLY consider those from the seconf
-- group 
commonUnsatConsLimit :: [BCons] -> [BCons] -> ([BCons], Bool)
commonUnsatConsLimit bcs1 bcs2 =
    let bbcs = allLeft bcs2
	com  = [ bc | (bc, bcs) <- bbcs, satBCs (bcs ++ bcs1) ]
    in  (com, unsatBCs com)

--------------------------------------------------------------------------------
-- interface to the min subset finding functions
-- These functions always find a set of minimal subsets, but these sets are not 
-- guaranteed to be complete! (depends on the algorithm used.)

minImplSubsets :: [BCons] -> BCons -> [[BCons]]
minImplSubsets bcs bc@(Eq (Var _) (Var _) _) = []
minImplSubsets bcs bc = oneMinImplSubset bc bcs

minUnsatSubsets :: [BCons] -> [[BCons]]
minUnsatSubsets bcs = oneMinUnsatSubset bcs

--------------------------------------------------------------------------------
-- first algorithm from paper - finds some non-overlapping minimal sets

oneMinUnsatSubset bcs   = [ oneMinSubset unSatBCs bcs ]

oneMinImplSubset bc bcs = [ oneMinSubset ((`entails`[bc])) bcs ]
--oneMinImplSubset bc bcs = [ oneMinSubsetInc ((`entails`[bc])) bcs ]

-- calls oneMinSubset repeatedly, accumulating one set of disjoint minimal
-- subsets (not necessarily unique. ie. there could be subsets it misses)
disjointMinSubsets :: BCPred -> [BCons] -> [[BCons]]
disjointMinSubsets p ds = findAll ds
    where findAll ds = let ms = oneMinSubset p ds
			   ns = filter (`notElem`ms) ds
		       in  if not (null ns) && p ns
			      then let mss = findAll ns
			           in  (ms : mss)
			      else [ms]

----------------------------------------
{-
-- finds one minimal subset of ds, with the property predicated by p
-- (assumes that the given set satisfies the predicate!)
oneMinSubset :: BCPred -> [BCons] -> [BCons]
oneMinSubset p ds = accumSet ds []
    where
	  -- accumulates all the unsatisfiable constraints 
	  -- (last ones found by initUnsat)
	  accumSet cs ms = let (e:es) = initSet cs ms
			       ms' = e : ms
			   in  if p ms'
				  then ms'
				  else accumSet es ms'
	   
	  -- returns the smallest prefix of list which is unsatisfiable, given
	  -- the initial set of constraints ms
	  initSet [] ms = ms
	  initSet (c:cs) ms
		| p (c:ms)  = c : ms
		| otherwise = initSet cs (c:ms)
-}

oneMinSubset = oneMinSubsetBCons

oneMinSubsetBCons :: BCPred -> [BCons] -> [BCons]
oneMinSubsetBCons p bcs = let p' xs = return (p xs) 
			  in  runIdentity $ oneMinSubsetGeneric p' bcs

oneMinSubsetMaybe :: ([a] -> Bool) -> [a] -> Maybe [a]
oneMinSubsetMaybe p bcs = let p' xs = return (p xs)
			  in  runIdentity $ oneMinSubsetGenericMaybe p' bcs

oneMinSubsetConstraint :: Pred Constraint -> Constraint -> Constraint
oneMinSubsetConstraint p c =
	let cs = [ C [uc] [] | uc <- consToUs c ] ++ 
		 [ C [] [bc] | bc <- consToBs c ]
	    p' = return . p . candList
	in  candList $ runIdentity $ oneMinSubsetGeneric p' cs

--mOneMinSubsetConstraint :: Monad m => (Constraint -> m Bool) -> 
--				       Constraint -> m Constraint
mOneMinSubsetConstraint p c = 
    let cs = [ C [uc] [] | uc <- consToUs c ] ++ 
	     [ C [] [bc] | bc <- consToBs c ]
	p' c = do b <- p (candList c)
		  return b
	
    in  do mc <- oneMinSubsetGenericMaybe p' cs
	   return (fmap candList mc)


-- finds one minimal subset of ds, with the property predicated by p
-- (assumes that the given set satisfies the predicate!)
oneMinSubsetGeneric :: (Show a, Monad m) => ([a] -> m Bool) -> [a] -> m [a]
oneMinSubsetGeneric p ds = accumSet ds []
    where
	  -- accumulates all the unsatisfiable constraints 
	  -- (last ones found by initUnsat)
	  accumSet cs ms= --trace ("\naccumSet\ncs: " ++ show cs ++ 
			  --		 "\nms: " ++ show ms) $
	  
			  (do (e:es) <- initSet cs ms
			      let ms' = e : ms
			      b <- p ms'
			      if b then return ms'
				   else accumSet es ms')
	   
	  -- returns the smallest prefix of list which is unsatisfiable, given
	  -- the initial set of constraints ms
	  initSet [] ms = trace "OUT OF CONSTRAINTS" $ return ms
	  initSet (c:cs) ms = do 
		b <- p (c:ms)
		if b then return (c:ms)
		     else initSet cs (c:ms)


oneMinSubsetGenericMaybe :: Monad m => ([a] -> m Bool) -> [a] -> m (Maybe [a])
oneMinSubsetGenericMaybe p ds = accumSet ds []
    where
	  -- accumulates all the unsatisfiable constraints 
	  -- (last ones found by initUnsat)
	  accumSet cs ms= do ees <- initSet cs ms
			     case ees of
				[] -> return Nothing
				(e:es) -> do let ms' = e : ms
					     b <- p ms'
					     if b then return (Just ms')
						  else accumSet es ms'
	   
	  -- returns the smallest prefix of list which is unsatisfiable, given
	  -- the initial set of constraints ms
	  initSet [] ms = return []
	  initSet (c:cs) ms = do 
		b <- p (c:ms)
		if b then return (c:ms)
		     else initSet cs (c:ms)



oneMinSubsetInc :: BCPredInc -> [BCons] -> [BCons]
oneMinSubsetInc p ds = accumSet ds newIncState
    where
	  -- accumulates all the unsatisfiable constraints 
	  -- (last ones found by initUnsat)
	  accumSet cs ms = let (e:es) = initSet cs [] ms
			       (ms', b) = p (ms, e)
			   in  if b then snd ms'
				    else accumSet es ms'
	   
	  -- returns the smallest prefix of list which is unsatisfiable, given
	  -- the initial set of constraints ms
	  initSet [] ds _ms = ds
	  initSet (c:cs) ds ms =
	    let (ms', b) = p (ms, c)
	    in  if b then c:ds
		     else initSet cs (c:ds) ms'


----------------------------------------

-- finds and returns minimal unsatisfiable subsets of bcs as well as a list of
-- the constraints which are uninvolved
minUnsatSubsetsAndRest :: [BCons] -> ([[BCons]], [BCons])
minUnsatSubsetsAndRest bcs = oneMinUnsatSubsetAndRestInc bcs

oneMinUnsatSubsetAndRest :: [BCons] -> ([[BCons]], [BCons])
oneMinUnsatSubsetAndRest bcs = 
    let (us, rs) = oneMinSubsetAndRest unSatBCs bcs
    in  ([us], rs)

-- finds one minimal subset as above, but also returns all the other
-- constraints as well (the ones not involved)
oneMinSubsetAndRest :: BCPred -> [BCons] -> ([BCons], [BCons])
oneMinSubsetAndRest p ds = accumSet ds [] []
    where
	  -- accumulates all the unsatisfiable constraints 
	  -- (last ones found by initUnsat)
	  accumSet cs ms as = let ((e:es), as') = initSet cs [] ms as 
			          ms' = e : ms
			      in  if p ms'
				     then (ms', es ++ as')
				     else accumSet es ms' as'
	   
	  -- returns the smallest prefix of list which is unsatisfiable, given
	  -- the initial set of constraints ms
	  -- as well as the left-overs that are uninvolved
	  initSet [] ds ms as = (ds, as)
	  initSet (c:cs) ds ms as
		| p (c:ms)  = (c:ds, cs ++ as)
		| otherwise = initSet cs (c:ds) (c:ms) as


oneMinUnsatSubsetAndRestInc :: [BCons] -> ([[BCons]], [BCons])
oneMinUnsatSubsetAndRestInc bcs = 
    let (us, rs) = oneMinSubsetAndRestInc unsatBCsInc bcs
    in  ([us], rs)


-- as above, but solves incrementally
oneMinSubsetAndRestInc :: BCPredInc -> [BCons] -> ([BCons], [BCons])
oneMinSubsetAndRestInc p ds = accumSet ds newIncState []
    where
	  -- accumulates all the unsatisfiable constraints 
	  -- (last ones found by initUnsat)
	  accumSet cs ms as = 
	    let ((e:es), as') = initSet cs [] ms as 
		(ms', b) = p (ms, e)
		in  if b then (snd ms', as' ++ es)
		         else accumSet es ms' as'
	   
	  -- returns the smallest prefix of list which is unsatisfiable, given
	  -- the initial set of constraints ms
	  -- as well as the left-overs that are uninvolved
	  -- cs: constraints yet to be included
	  -- ds: constraints originally from cs which have been included
	  -- ms: ds combined with original "definitely in" constraints (subst)
	  -- as: constraints not used
	  initSet [] ds ms as = (ds, as)
	  initSet (c:cs) ds ms as =
	    let (ms', b) = p (ms, c)
	    in  if b then (c:ds, cs ++ as)
		     else initSet cs (c:ds) ms' as


--------------------------------------------------------------------------------
-- algorithm for finding all min. sets
-- (actually, finds some of their supersets too)

-- first list are the "definitely-ins", we choose constraints from the second
allMinUnsatSubsets :: [BCons] -> [BCons] -> [[BCons]]
allMinUnsatSubsets f0 [] | unsatBCs f0	= [f0]
			 | otherwise	= []
allMinUnsatSubsets f0 c0@(c:cs) = tilUnsat f0 c0
    where 
	tilUnsat f c 
	    | satBCs f && not (null c) = tilUnsat ((head c):f) (tail c)
	    | otherwise = 
		if unsatBCs f
		    then let k   = head f
			     f0' = k : f0
			     c0' = delete k c0
			 in  if unsatBCs f0'
				then [f0'] ++ allMinUnsatSubsets f0 c0'
				else allMinUnsatSubsets f0  c0' ++
				     allMinUnsatSubsets f0' c0'
		    else []

--------------------------------------------------------------------------------
-- finds all minimal subsets by removing a constraint at a time
-- (cuts out if the set is sat, though - doesn't consider its subsets)
-- the standard dynamic programming approach (apparently)

dpMinImplSubsets bcs bc = dpMinSubsets (\bs -> not (bs `entailType` bc)) bcs 
dpMinUnsatSubsets = elimSupersets . dpMinSubsets satBCs
		    -- dpMinSubsets satBCs

dpMinSubsets :: BCPred -> [BCons] -> [[BCons]]
dpMinSubsets p r = findAll [] r
    where 
	findAll s r = 
	    let sr = s ++ r
	    in  if p sr 
		   then []
		   else if null r
			   then [s]
			   else let c  = head r
				    s1 = c : s
				    r1 = tail r
				    a1 = findAll s1 r1 
				    a2 = findAll s  r1 
				    as = a1 ++ a2
				in  if null as
				       then [sr]
				       else as -- elimSupersets as


--------------------------------------------------------------------------------
-- finds all min. subsets by trying all possible subsets, then eliminating 
-- supersets from the result

psMinImplSubsets bcs bc = let gcs = groupBCons bcs
			  in  psMinSubsets (`entailType`bc) gcs

psMinUnsatSubsets bcs = let gcs = groupBCons bcs
			in  psMinSubsets unsatBCs gcs
			    
psMinSubsets :: BCPred -> [GBCons] -> [[BCons]]
psMinSubsets p c = let cs1 = tail $ powerset (map snd c)
		       cs2 = filter (p . concat) cs1
		   in  elimSupersets (map concat cs2)

--------------------------------------------------------------------------------

-- given an unsatisfiable set of constraints, finds all maximal satisfiable
-- subsets
-- 1) finds all minimal unsatisfiable subsets, Us
-- 2) "neuter" each such set as it appears in the original, to make the
--    original set of constraints satisfiable
--maxSatisfiable :: [BCons] -> [[BCons]]
maxSatisfiable bcs = 
    let -- min. unsat. subsets
	us = dpMinUnsatSubsets bcs
	
	-- constraints to remove from bcs
	us' = makeUnique us
	rs  = comb (filter (not.null) us) -- us')
    
	sats = map (bcs`without`) rs
    in  sats

-- given a list of alternatives (lists), generates all lists with each possible
-- combination of alternatives. 
-- eg. comb [[1,2],[3,4]] ==> [[1,3],[1,4],[2,3],[2,4]]
comb :: [[a]] -> [[a]]
comb = foldl cons [[]]		-- foldr (flip cons) [[]] 
    where cons yss = concatMap (\x -> map (x:) yss)

-- removes the minimal number of elements from each list such that no list
-- contains any elements of another
makeUnique :: Eq a => [[a]] -> [[a]]
makeUnique [] = []
makeUnique (xs:xss) = xs : makeUnique (delAll xs xss)

    where 
	delAll :: Eq a => [a] -> [[a]] -> [[a]]
	delAll xs [] = []
	delAll xs (ys:yss) = del xs ys : delAll xs yss

	del :: Eq a => [a] -> [a] -> [a]
	del [] ys = ys
	del (x:xs) ys = del xs (delete x ys)

	  
--------------------------------------------------------------------------------

{-
-- finds the smallest t <= t2 such that unify(t1, t2) fails
normalisedErrorType :: Int -> (Type, Type) -> (Int, Type)
normalisedErrorType n ts = case ts of
    (TyCon n1, t2@(TyCon n2)) | n1 /= n2 -> (n, t2)
    (TApp t1 t2, TApp s1 s2) -> let (n1,q1) = normalisedErrorType n  (t1,s1)
				    (n2,q2) = normalisedErrorType n1 (t2,s2)
				in  if isVar q1 && isVar q2
				    then (n2+1, newVar n2)
				    else (n2, TApp q1 q2)
    

    where newVar n = Var (name (show n) n, unkIx n)
-}

-- finds the smallest t <= t1 such that unify(t1, t2) fails
-- (Int is for generating new variables)
innerr :: Int -> (Type, Type) -> Type
innerr n ts = thd3 (strip n idSubst ts)
    where
	new n = Var (name ("q" ++ show n) (fromIntegral n), n)

	strip n s ts@(t1,t2) | isSucc s' = (n+1, unSucc s', new n)
	    where s' = unify [] s [ts] 
	    
	strip n s (TApp t1 t2, TApp q1 q2) =
	    let (n1, s1, t1') = strip n  s  (t1,q1)
		(n2, s2, t2') = strip n1 s1 (t2,q2)
	    in  (n2, s2, TApp t1' t2')
{-	
	strip n s (TApp t1 t2, _) = (n, s, left t1)
	    where left (TApp t _) = left t
		  left t = t
-}	
	strip n s (TApp _ _, _) = (n+2,s,TApp (new n) (new (n+1)))
	strip n s (t1,_) = (n, s, t1)

test = innerr 1

--------------------------------------------------------------------------------


n1 = Name ("n1", 1)
n2 = Name ("n2", 2)
n3 = Name ("n3", 3)
n4 = Name ("n4", 4)
n5 = Name ("n5", 5)
n6 = Name ("n6", 6)
n7 = Name ("n7", 7)
n8 = Name ("n8", 8)
n9 = Name ("n9", 9)

v1 = (n1, 1)
v2 = (n2, 2)
v3 = (n3, 3)
v4 = (n4, 4) 
v5 = (n5, 5)
v6 = (n6, 6)
v7 = (n7, 7)
v8 = (n8, 8)
v9 = (n9, 9)

arr = TyCon (Name ("->", 0))
lst = TyCon (Name ("[]", (-1)))
tup = TyCon (Name ("(,)", (-2)))
int = TyCon (Name ("Int", (-3)))
float = TyCon (Name ("Float", (-4)))
char  = TyCon (Name ("Char", (-5)))

a t1 t2 = TApp (TApp arr t1) t2
p t1 t2 = TApp (TApp tup t1) t2
l t = TApp lst t

t1 = Var v1
t2 = Var v2
t3 = Var v3
t4 = Var v4
t5 = Var v5
t6 = Var v6
t7 = Var v7
t8 = Var v8
t9 = Var v9

f1 = Eq t1 t2 (J [1])
f2 = Eq t2 int (J [2])
f3 = Eq t1 t4 (J [3])
f5 = Eq t1 t1 (J [5])

f4 = Eq int t1 (J [4])

{-
f1 = Eq t1 (p t2 t3) (J [1])
f2 = Eq t1 (p t4 t4) (J [2])
f3 = Eq t2 (l t5) (J [3])
f4 = Eq t3 (l t6) (J [4])
f5 = Eq t5 int  (J [5])
f6 = Eq t6 char (J [6])
f7 = Eq t4 float (J [7])
-}

