--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- functions for reasoning about resulting constraint stores:
--	- minimal unsatisfiable subsets
--	- minimal entailment sets

module Constraint (module Constraint, module UCons, module Builtin)
where

import Types
import UCons	hiding (Index)	    -- it's a Chr (solver) thing only
import Builtin
import Justify
import Pretty
import List



-- Constraints

data Constraint = C [UCons] [BCons] -- deriving Show

instance Eq Constraint where
    (==) (C us1 bs1) (C us2 bs2) = us1 == us2 && bs1 == bs2

instance Justifiable Constraint where
    unJust (C us1 bs1) = unJust us1 ++ unJust bs1
    reJust j (C us1 bs1) = C (reJust j us1) (reJust j bs1)

--instance ShowAnn Constraint where
--    showAnn (C ucs bcs) = "C " ++ showAnn ucs ++ showAnn bcs

instance PPrint Constraint where
    pprint (C ucs bcs) = do ducs <- catListWith "," ucs
			    dbcs <- catListWith "," bcs
			    if not (null ucs) && not (null bcs)
			       then return (ducs <> text "," <> dbcs)
			       else return (ducs <> dbcs)

instance Show Constraint where show c = pprintPlain c

instance TC Constraint where
   apply s (C ucs bcs) = C (apply s ucs) (apply s bcs)
   fv (C ucs bcs) = (fv ucs) ++ (fv bcs)
   occurs v (C ucs bcs) = (occurs v ucs) || (occurs v bcs)
   rename c = do s <- renameSubst (fv c)
                 return (apply s c)
   format (C ucs bcs) = let u = (format ucs)
                            b = (format bcs)
                        in if u == []
                           then b
                           else if b == []
				then u
			        else "("++u++","++b++")"

trueCons, falseCons :: Constraint
trueCons = C [] []
falseCons = C [] [njEq (TyCon trueName) (TyCon falseName)]

-- also true, but distinct from the standard trueCons
trueConsAlt :: Constraint
trueConsAlt = C [] [njEq (TyCon trueName) (TyCon trueName)]

addCons :: Constraint -> Constraint -> Constraint
addCons (C u1 b1) (C u2 b2) = C (u1++u2) (b1++b2)

bToCons :: BCons -> Constraint
bToCons b = C [] [b]

uToCons :: UCons -> Constraint
uToCons u = C [u] []

consToUs :: Constraint -> [UCons]
consToUs (C us _) = us

consToBs :: Constraint -> [BCons]
consToBs (C _ bs) = bs


bsToCons bs = C [] bs
usToCons us = C us []
cToUCons = consToUs
cToBCons = consToBs


cand (C u1 b1) (C u2 b2) = C (u1++u2) (b1++b2)
candList cs = foldr cand trueCons cs

remRedundantCons :: Constraint -> Constraint
remRedundantCons (C us bs) = C (removedups2 us) (remRedundantBCons bs)
