--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- Justifications

module Justify where

import Terminal (just, commaList)
import Ann_InputAST (Loc, Index(..), unkIx, noLoc)
import Misc (sameSet, subset)

import List (nub)

newtype Just = J [Loc]

noJust = J []

isNoJust (J ls) = null ls

nullJust (J []) = True
nullJust _	= False

inJust l (J ls)	  = l `elem` ls
-- unJust (J ls)	  = ls
ixToJust NoIx	    = J []
ixToJust (Ix l _ _) = J [l]
ixsToJusts	  = map ixToJust
locToJust l	  = J [fromIntegral l]

justToLoc (J ls) = if null ls then noLoc else head ls

justToIx (J (l:_))  = unkIx l
justToIx _	    = error "Justify.justToIx: noJust!"

-- addJust (J ls1) (J ls2) = J (nub (ls1 ++ ls2))
addJust (J ls1) (J ls2) = J (ls1 ++ ls2)
catJust (J ls1) (J ls2) = J (ls1 ++ ls2)
subJust (J ls1) (J ls2) = ls1 `subset` ls2
nubJust (J ls) = J (nub ls)

class Justifiable a where
    unJust :: a -> [Loc]
    reJust :: Just -> a -> a

instance (Justifiable a, Justifiable b) => Justifiable (a, b) where
    unJust (a, b) = unJust a ++ unJust b
    reJust j (a, b) = (reJust j a, reJust j b)

instance Justifiable Just where
    unJust (J ls) = ls
    reJust = const

instance Justifiable a => Justifiable [a] where
    unJust xs = nub (concatMap unJust xs)
    reJust j xs = map (reJust j) xs
    
instance Justifiable a => Justifiable (Maybe a) where
    unJust Nothing  = []
    unJust (Just x) = unJust x

    reJust j Nothing  = Nothing
    reJust j (Just x) = Just (reJust j x) 
    

instance Show Just where
    
    show (J [])	    = "*"
    show (J [l])    = show l
    show (J ls)	    = commaList ls

instance Eq Just where

    (J l1) == (J l2) = l1 `sameSet` l2


-- describes policies for generation of justifications on constraints freshly
-- added to the store
-- OrigJust:  the original justification (as given to the constraint when
--	      it was first dreated during rule generation) is used
-- NewJust:   the new constraint's justification is given by the sum of
--	      justifications on the constraints in the store responsible for
--	      the rule which adds the new constraints having fired
data JustProp = GlobalJust | LocalJust  deriving Eq

instance Show JustProp where
    show GlobalJust = "global"
    show LocalJust  = "local"

isLocalJust = (==) LocalJust  
isGlobalJust= (==) GlobalJust 
isNewJust   = isLocalJust
isOrigJust  = isGlobalJust


-- justification given to administrative constraints
adminLoc = -2
adminJust = J [adminLoc]
