--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- defines control flags which are used to control various
-- aspects of the running Chameleon system

module OptFlags (Flag(..), verbose, noCkUnambigFlag, noChecks, compileFlag,
		 outputFlag, jpGlobalFlag)
where

type SLvl = Int

data Flag = InitEnv	    -- display "initial environment"
	  | IntSyntax	    -- display internal syntax (for parsed code)
	  | NoCkRR	    -- ignore range-restrictedness check
	  | NoCkConf	    -- ignore confluence check
	  | NoCkTerm	    -- ignore termination check
	  | NoCkUnambig	    -- ignore unambiguity check
	  | NoChecks	    -- ignore all checks 
	  | JustChecks	    -- only perform chr checks
	  | H98Term	    -- use Haskell98 termination criteria
	  | Complete	    -- attempt to complete nonconfluent chrs
	  | Debug	    -- invoke type debugger
	  | CkSimpOverlap   -- check for overlapping simp rules
	  | Version	    -- display version
	  | Verbose	    -- report more information
	  | Usage	    -- display usage/help
	  | Compile	    -- if present, perform evidence translation
	  | SolverOpt SLvl  -- initial solver option level
	  | JPGlobal	    -- global just policy
	  | Output String   -- output file
    deriving (Show, Eq)

verbose	    fs = Verbose `elem` fs
noChecks    fs = NoChecks `elem` fs
noCkUnambigFlag fs = NoCkUnambig `elem` fs
compileFlag     fs = Compile `elem` fs
jpGlobalFlag    fs = JPGlobal `elem` fs

-- checks for the output flag
outputFlag :: [Flag] -> Maybe String
outputFlag fs = 
    let strs = concat $ map (\f -> case f of Output s -> [s]; _ -> []) fs
    in  if null strs then Nothing
		     else Just (head strs)
        
