--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

-- contains code for parsing (and handling of trivial) command-line options

module Options (getOpts, Flag(..), defaults)
where

import OptFlags
import qualified Version

import GetOpt
import IO
import System
import Monad
import List


options :: [OptDescr Flag]
options =
    [ Option ['c'] ["complete"]	  (NoArg Complete) "complete rules"
    , Option []	   ["h98-term"]	  (NoArg H98Term)  "use Haskell 98 termination conditions"
    , Option []	   ["ck-simp-overlap"] (NoArg CkSimpOverlap) 
				  "reject programs with overlapping simp. rules"
    , Option []    ["no-rr"]	  (NoArg NoCkRR) "don't do range-restrictedness check"
    , Option []    ["no-conf"]	  (NoArg NoCkConf) "don't do confluence check"
    , Option []    ["no-term"]	  (NoArg NoCkTerm) "don't do termination check"
    , Option []	   ["no-unambig"] (NoArg NoCkUnambig) "don't do unambiguity check"
    , Option []    ["no-checks"]  (NoArg NoChecks) "don't perform any checks"
    , Option ['s'] ["solver"]	  (ReqArg (SolverOpt . read) "lvl") "init solver mode"
    , Option ['g'] ["jp-global"]  (NoArg JPGlobal) "global justification policy"
    , Option ['d'] ["debugger"]   (NoArg Debug)   "invoke debugger"
    , Option ['o'] []		  (ReqArg Output "file") "output file"
    , Option []    ["verbose"]	  (NoArg Verbose) "report more information"
    , Option ['v'] ["version"]    (NoArg Version) "show release version"
    , Option ['h'] ["help"]	  (NoArg Usage) "display help (this text)"
    ]

defaults :: [Flag]
defaults = []


normFlags :: [Flag] -> [Flag]
normFlags fs
    | NoChecks `elem` fs = comp ([NoCkRR, NoCkConf, NoCkTerm] `union` fs)
    | otherwise		 = comp fs
    where comp fs | Debug `notElem` fs = Compile : fs
		  | otherwise	       = fs

getOpts :: IO ([Flag],[String])
getOpts =
    do  argv <- getArgs
        case (getOpt Permute options argv) of
            (o,n,[])    -> do handleImmediateFlags o ; return (normFlags o,n)
            (_,_,er)    -> do putStr $ concat er ; help ; exitFailure

--------------------------------------------------------------------------------

-- performs opperations for flags that need to be dealt with immediately
handleImmediateFlags :: [Flag] -> IO ()
handleImmediateFlags fs = handleVersion fs >> handleUsage fs

handleUsage, handleVersion :: [Flag] -> IO ()
handleUsage fs   = when (Usage `elem` fs) (usage >> exitWith ExitSuccess)
handleVersion fs = when (Version `elem` fs) (version >> exitWith ExitSuccess)

version = putStrLn $ Version.str 

help    = putStr $ "try `--help' for basic usage information\n"

usage   = putStr $ usageInfo usageHeader options
    where   usageHeader = "Usage:\n\tchameleon [options] <file>\n"
