--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


module TDExplain (explainLocType, explainLocTypeOpt, 
		  explainExpType, explainRefType)
where

import TD_Monad
import Core
-- import Chr
import TD_General
import Ann_InputAST
import Justify
import Misc
import TDInference
import ConsReason
import TD_Translate (transDef, transTSchm)

import ASTs (lookupVarTree, vtProg, VarRef(..))

import CHRChecks (entail2, entail3)	-- MOVE entail2 OUT OF THERE!

import Monad
import Maybe
import List

--import IOExts


--------------------------------------------------------------------------------

explainRefType :: VarRef -> Ann TSchm -> TD ()
explainRefType vr ts = do 
    p  <- getProg
    let vt = vtProg p
	midl = lookupVarTree vr vt
    case midl of
	Right id    -> puts ("Undefined variable `" ++ id ++ "'")
	Left (id,l) -> explainLocType l l ts >> return ()


--------------------------------------------------------------------------------
		   
-- copied structure from inferTypeExp
explainExpType :: Ann Exp -> Ann TSchm -> TD ()
explainExpType exp tschm = do
    jp <- getJProp
    case item exp of
     (VarExp id) -> do -- lookup the corresponding def, infer its type
	(Prog decs) <- getProg
	chrs <- getCHRs
	let lis  = valDecsToLocIds decs
	    locs = [ l | (l, id') <- lis, id == id' ]
	    loc  = head locs
	if null locs
	   then puts ("`" ++ id ++ "' undefined!")
	   else	-- inferTypeLoc loc (True, True, False) jp
		explainLocType loc loc tschm >> return ()
		   
     
     _		 -> do -- convert to a def, then a chr, then explain type
        n1 <- next
	n2 <- next
	n3 <- next
	n4 <- next
	let id   = "(" ++ show n3 ++ ")"
	    def  = Def (Ann (unkIx (fromIntegral n1)) id) [] (RHS1 exp) []
	    adef = Ann (unkIx (fromIntegral n2)) def
	    dec  = ValDec [adef]
	    adec = Ann (unkIx (fromIntegral n3)) dec
	prog <- getProg
        chrs <- getCHRs
	env  <- getEnv

	let Prog ds = prog
            prog' = Prog (adec:ds)
	res <- doTR $ transDef env adef
	case res of
	 Err e -> puts e
	 Succ chr -> do
	    let chrs' = chr : chrs 
		loc   = ixNum (ann exp)
    
	    setProg prog'
	    setCHRs chrs'
    
	    explainLocType loc loc tschm
	    setProg prog
	    setCHRs chrs


-- see below
explainLocType = explainLocTypeOpt True

-- explain why the type scheme inferred for location, l1, given constraints 
-- within l2, is an instance of ts
-- takes an additional flag, if False, will not provide a textual explanation
explainLocTypeOpt :: Bool -> Loc -> Loc -> Ann TSchm -> TD [Loc]
explainLocTypeOpt disp l1 l2 ts = do
    jp <- getJProp
    
    -- always use the debugging solver (mode 2)
    sopt <- getSolvOpt
    setSolvOpt SDebug
    inferTypeLoc l2 (False, True, False) jp
    setSolvOpt sopt
    
    res  <- lastResult
    if noResult res
       then puts ("location " ++ show l2 ++ " doesn't exist") >> retNone
       else do	
	prog <- getProg
	chrs <- getCHRs
	let ix  = unkIx l1 
	    tcs = chrWithVarIx (ixLoc ix) chrs
	case tcs of
	     []	-> puts ("location " ++ show l1 ++ " not found") >> retNone
	     ((tv,chr,rest):_) -> do
		env <- getEnv
		n   <- next
		tres <- doTR (transTSchm env ts)
		case tres of
		     Err e -> puts e >> retNone
		     Succ (typ, ctxt) -> do
			let u s = error ("explainLocType error:" ++ show s)
			    (lt1, _k_lts, _nk_lts, cons, _scons) = 
				if isConflict res 
				   then let (_,l,_,c) = unConflict res
				        in  (l,u 2,u 3,c,u 4)
				   else let (v,w,x,(y,z),_) = unResult res
				        in  (fst3 v,w,x,y,z)
			    lts = [lt1] -- : k_lts ++ nk_lts
			    mtv = lookup l1 lts
			    tv  = fromJust mtv   -- potential bug

			    sigcons = bToCons (Eq (Var tv) typ noJust) `cand`
				      ctxt
			    
--			ment <- doTP $ entail3 chrs (fv cons) (cons,sigcons)

			let pred = \c -> do (b,_) <- entail3 chrs (fv cons) (c,sigcons)
					    return b
{-
			puts ("sigcons: " ++ show sigcons)
			puts ("   cons: " ++ show cons)
			puts ("     vs: " ++ show (fv cons))
-}
			-- check if the queried type is consistent with
			-- the inferred type
			let typeOkay = satBCs (consToBs (sigcons `cand` cons))

			    noExp = do puts "This type cannot be explained."
				       return []

			if typeOkay
			 then do min<- mOneMinSubsetConstraint pred cons
				 -- let min = unSucc mmin
				 if isJust min
				    then do let min' = fromJust min
						ls = nub (unJust min')
--					    puts ("min: " ++ show min')
--					    puts ("ls: " ++ show ls)
					    highlightConf ls
					    return ls
				    else noExp
			 else noExp
{-			 
 			
			    -- type explanations
			    tents = map nub $ expTyp cons tv typ
			    -- context explanations
			    (cents, umcs) = expCtxt cons ctxt
			    -- HACK (fix expCtxt)
			    cents'= [ ((f, s), ls') | ((f, s), ls) <- cents,
					let ls' = if null ls then [unJust f]
							     else ls ]
			    
			    match = typMatch scons tv typ
			    
			do puts ("\nlt1: " ++ show lt1 ++
				 "\ncons: " ++ show cons ++
				 "\nscons: " ++ show scons ++
				 "\nctxt: " ++ show ctxt ++
				 "\ntents: " ++ show tents ++
				 "\ncents: " ++ show cents ++
				 "\ncents': " ++ show cents' ++ 
				 "\numcs: " ++ show umcs)

			    
			if isConflict res
			 then dispExp False ((error "XXX")::Type) ctxt [] 
				      cents' umcs >> retNone
				      
			 else if isNothing mtv 
			      then puts ("could not find " ++ show l1 ++ 
					 " within " ++ show l2) >> retNone
			      else if not match
				   then puts "type mismatch" >> retNone
				   else do 
					 dispExp True typ ctxt tents cents' umcs
					 return (nub (concat tents ++ 
						 concat (concatMap snd cents)))
-}    
  where   
   
    retNone = return []
{-  
    dispExp = if disp then dispExp2
		      else \_ _ _ _ _ _ -> return ()

    -- display the explanation
    dispExp2 :: Bool -> Type -> Constraint -> [[Loc]] -> 
	       [((UCons,UCons),[[Loc]])] -> [UCons] -> TD ()
    dispExp2 b typ ctxt tents cents umcs =
	if null tents && ctxt == trueCons
	   then puts "trivial"
	   else do when (b && not (null tents))
		        (do puts ("type of form: " ++ 
				    show (presentableVars typ))
			    puts (" due to " ++ show (map sort tents))
			    mapM_ report tents)
		   when (length (consToUs ctxt) > length cents)
			(puts ("supplied context cannot be fully " ++
			       "explained"))
		   when (umcs /= [])
			(puts ("following supplied constraints " ++
			       "cannot be matched with inferred " ++
			       "constraints:\n\t" ++ unwords 
			        (map show (presentableVars umcs))))
		   when (ctxt /= trueCons) (mapM_ dispCtxt cents)

    report ls = do highlightConf ls 
		   puts ""

    dispCtxt (cons, exp) = do
	let cstr = show (presentableVars (snd cons))
	if null exp 
	   then puts (cstr ++ " trivial")
	   else do puts (cstr ++ " due to " ++ show exp)
		   mapM_ report exp


    -- if True, then the inferred type is an instance of
    -- the queried type
    typMatch scons tv t =
	let bcs = consToBs scons
	    phi = unSucc (unify [] idSubst (buildUP bcs))
	    tv' = apply phi (Var tv)
	in  matchTypes [t] [tv']
	
	  
    -- explain the type
    expTyp cons tv typ =
	let bc   = Eq (Var tv) typ noJust
	    ents = entailSubsetExp cons bc
	in  elimSupersets ents

    entCons cons bc = elimSupersets (entailSubsetExp cons bc)

    -- explain the context
    -- matches constraints from the inferred context (cons), with those
    -- from the supplied context (ctxt), and for each successful match,
    -- finds a list of locations which imply it
    -- also returns a list of the supplied user constraints which could
    -- not be matched against inferred ucons
    expCtxt :: Constraint -> Constraint -> 
	       ([((UCons, UCons), [[Loc]])], [UCons])
    expCtxt cons ctxt =
	let sucs = cToUCons ctxt
	    pucs = pairCons sucs (cToUCons cons)
	    exps = map (flip expCons cons) pucs
	    cexps  = zip pucs exps
	    cexps' = [ pls | pls@(_, ls) <- cexps, ls /= [] ]

	    umcs  = [ c | c <- sucs, c `notElem` map (snd . fst) cexps']
	in trace ("\nCONS: " ++ show cons ++ 
		  "\nSUCS: " ++ show sucs ++
                  "\nPUCS: " ++ show pucs ++
                  "\nEXPS: " ++ show exps) (cexps', umcs)

    -- generates all possible pairings of constraints from cs with
    -- those from ds
    -- constraints are paired if they have the same name
    pairCons :: [UCons] -> [UCons] -> [(UCons, UCons)]
    pairCons []     _  = []
    pairCons (c:cs) ds = 
	let ucss = [ (d, c) | d <- ds, uconsPSym c == uconsPSym d ]
	in  (trace $ "pairCons: " ++ show (map uconsPSym (c:cs), map uconsPSym ds)) 
		ucss ++ pairCons cs ds

    -- takes a pair of UCons (unsimplified, simplified) and all the other
    -- constraints, and returns an explanation for the type being 
    -- predicated
    -- returns the empty list if the type in the query is not implied by
    -- the given constraint 
    expCons :: (UCons, UCons) -> Constraint -> [[Loc]]
    expCons (uc, suc) cons =
	let js    = ucIx uc
	    typ1  = head (ucTypes uc)
	    typ2  = head (ucTypes suc)
	    bc    = Eq typ1 typ2 noJust
	    
	    ments = consToBs cons `subsetEntails` [bc]
	    ents  = isJust ments
	    cons' = bsToCons (fromJust ments)
	    texps = entCons cons' bc
	    texps' = trace ("texps:\tcons: " ++ show cons' ++ 
				  "\tbc: " ++ show bc ++
				  "\texps: " ++ show texps) texps
	in  if not ents
	    then []
	    else map ((unJust js)`union`) texps'
-}

--------------------------------------------------------------------------------

{-
-- finds a minimal subset of second constraint which implies the first
-- uses entail2
-- use a generic version of ConsReason.oneMinSubset which we can use to find 
-- subsets of a Constraint
--explain :: Constraint -> Constraint -> TD Constraint
explain c1 c2 = do
    chrs <- getCHRs
    let p = \c -> do (b,_) <- entail2 chrs (fv c1) (c, c1)	-- wrong way around?
		     return b
    min <- doTP (oneMinSubsetConstraint p c2)
    return (unSucc min)
-}

{-
uc t = UC 1 (name "C" 1) [t] noJust

t1 = TyCon (name "X" 2)
t2 = Var  (name "a" 3, 4)
-}
