--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains functions to determine if, given a set of constraints, a given rule
-- will fire; and the minimal subset of the eq. constraints for which it will
-- do so

module TD_RuleMatch
where

import Core
import ConsReason
import Ann_InputAST (Loc, noLoc)
import Justify
import Misc

import List
import Maybe
import IOExts


matchable :: [(Type, Type)] -> [BCons] -> Bool
matchable tss bcs = 
    let teqs =  [ Eq t2 t1 (locToJust noLoc) | (t1, t2) <- tss ]
    in  bcs `entails2` (map (\(a,b)->(b,a))) tss -- all (bcs`entails2`) (map (:[]) teqs)


-- if flag is True, then just find if a match exists and return no
-- justification for it, else attempt to find the minimal implicants
-- NOTE: assumption is that first list come from the rule (second from store)
matchRule :: Bool -> Subst -> [Type] -> [Type] -> [BCons] -> Maybe [Loc]
matchRule b s ts1 ts2 bcs = 
	    let sameLength = length ts1 == length ts2
		mr = matchRule'
		typesMatch = isJust mr
	    in  if sameLength && typesMatch
		then mr
		else Nothing
    where 
	
	-- automatically accept the match if the head of the rule contains
	-- either:
	-- 1) a variable
	-- 2) a tuple of variables, and the matching constraint also contains a
	--    tuple of the same length
	autoAccept = 
	    and [ okay | (t1, t2) <- zip ts1 ts2,
			let var  = isVar t1
			    mtup = do tt1 <- tupleTypes t1
				      tt2 <- tupleTypes t2
				      return (length tt1 == length tt2)
			    tup  = isJust mtup && fromJust mtup
			    okay = var || tup ]
	
	matchRule'
	    | autoAccept = Just []
	    | matchTypes ts1 (apply s ts2) = Just $
		 if b then []
		 else let p bcs = let s= unSucc (unify [] idSubst (buildUP bcs))
				  in  matchTypes ts1 (apply s ts2) 
		          res = oneMinSubset p bcs
	    
		      in nub (unJust res)
		    
	    | otherwise = Nothing
		
{-
	    | autoAccept = Just []
	    | b		 = if matchTypes ts1 ts2 then Just [] else Nothing
	    | otherwise  = 
		let p bcs = let s = unSucc (unify [] idSubst (buildUP bcs))
			    in matchTypes ts1 (apply s ts2) 
		    res = oneMinSubset p bcs
	    
		    ls  = nub (unJust res)
		    mmgu = unify [] idSubst (buildUP bcs)
		    mgu  = unSucc mmgu
		in  -- if types can never match, just fail
		    if not (matchTypes ts1 (apply mgu ts2))
		    then Nothing
		    else Just ls
-}

ruleMatches :: Maybe [Loc] -> Bool
ruleMatches Nothing  = False
ruleMatches (Just _) = True


matchJust :: Maybe [Loc] -> Just
matchJust (Just ls) = J (filter (/=noLoc) ls)

