--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains functions which make use of extended terminal i/o features to do
-- things like displaying coloured text... use of colour makes annotated output
-- significantly easier to read

-- Also contains a number of other general-purpose text formatting functions
-- (mostly list formatting)

module Terminal (normal, ident, keyword, unknown, passed, failed, just,
		 commaList, Colour(..), bg, adjacentColour, plural,
		 hasBGEsc, uncommas, unsemis, linify, defaultBG )
where

import Prelude hiding (fail)
-- import TD_ShowAnn (ShowAnn, showAnn)


-------------
-- Colours

data Colour = Black | Red | Green | Yellow | Blue | Magenta | Cyan 
	    | White | Default	    deriving Eq


--adjacentColour Red     = Magenta
----adjacentColour Magenta = Blue
----adjacentColour Blue    = Magenta
adjacentColour _       = Red

defaultBG = bg Default

--------------------------------
-- Terminal control sequences

fg Black   = "\ESC[30m"
fg Red     = "\ESC[31m"
fg Green   = "\ESC[32m"
fg Yellow  = "\ESC[33m"
fg Blue    = "\ESC[34m"
fg Magenta = "\ESC[35m"
fg Cyan    = "\ESC[36m"
fg White   = "\ESC[37m"
fg Default = "\ESC[39m"

bg Black   = "\ESC[40m"
bg Red     = "\ESC[41m"
bg Green   = "\ESC[42m"
bg Yellow  = "\ESC[43m"
bg Blue    = "\ESC[44m"
bg Magenta = "\ESC[45m"
bg Cyan    = "\ESC[46m"
bg White   = "\ESC[47m"
bg Default = "\ESC[49m"


-----------------------------
-- Standard Colour Schemes

normal    s = defaultTerm ++ s

ident     s = yellow s ++ defaultTerm
keyword	  s = green s ++ defaultTerm

unknown	  s = (bold s) ++ defaultTerm
passed	  s = ((bold.blue) s) ++ defaultTerm
failed	  s = ((bold.red) s) ++ defaultTerm

just	  s = unknown s
--diag	  s = failed s


yellow s = fg Yellow ++ s
green  s = fg Green  ++ s
red    s = fg Red    ++ s
blue   s = fg Blue   ++ s

bold   s = "\ESC[1m" ++ s

defaultTerm  = "\ESC[0m"



--------------------------------------------------------------------------------
-- List formatting functions

commaList :: Show a => [a] -> String
commaList []	 = ""
commaList [x]	 = show x
commaList (x:xs) = show x ++ ", " ++ commaList xs


uncommas []     = ""
uncommas [x]    = x
uncommas (x:xs) = x ++ "," ++ uncommas xs

unsemis []     = ""
unsemis [x]    = x
unsemis (x:xs) = x ++ ";" ++ unsemis xs


-- just like unlines, but doesn't place a newline after the last line
linify []     = ""
linify [x]    = x
linify (x:xs) = x ++ "\n" ++ linify xs

plural [x] str = str
plural _   str = str ++ "s"


--------------------------------------------------------------------------------
-- misc.

-- True if the given string contains escape characters which change the
-- background colour
hasBGEsc :: String -> Bool
hasBGEsc ('\ESC':'[':'4':c:'m':cs) 
    | c /= '9' && c /= '0' = True
    | otherwise = hasBGEsc cs
hasBGEsc "" = False
hasBGEsc (c:cs) = hasBGEsc cs
