--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------

module Tokens where

import Misc (SrcPos)

data Token  = TId    (String, SrcPos)
	    | TConId (String, SrcPos)
	    | TIId   (String, SrcPos)
	    | TInt   (Integer, String)
	    | TFloat (Double, String)
	    | TChar Char
	    | TStr String
	    | TRule  
	    | TOverload
	    | TInstance
	    | THConstraint
	    | TExtern
	    | TInterface
	    | TData
	    | TSemicolon
	    | TLBrace
	    | TRBrace
	    | TEquals
	    | TBackslash
	    | TArrow
	    | TIf
	    | TThen
	    | TElse
	    | TLet
	    | TIn
	    | TCase
	    | TOf
	    | TWhere
	    | TLParen
	    | TRParen
	    | TLBracket
	    | TRBracket
	    | TCont
	    | TUnderscore
	    | TImpl
	    | TPropArrow
	    | TSimpArrow
	    | TComma
	    | TBar
	    | TLArrow
	    | TAnn  SrcPos
	    | TEAnn SrcPos
	    | TTrue
	    | TFalse
	    | TAsterisk
	    | TQuery
	    

	    | TLayoutBlock Int
	    | TLayoutFirst Int
	    
	    | Teof
	      deriving Eq 


instance Show Token where
    show tok = case tok of
	TId p@(string, n)	-> string
	TConId p@(string, n)	-> string
	TIId p@(string, n)	-> string
	TInt (_,str)	-> str
	TFloat (_,str)	-> str
	TChar char	-> show char
	TStr string	-> string
	TRule		-> "rule" 
	TOverload	-> "overload"
	TInstance	-> "instance"
	THConstraint	-> "hconstraint"
	TExtern		-> "extern"
	TInterface	-> "interface"
	TData		-> "data"
	TSemicolon	-> ";"
	TLBrace		-> "{"
	TRBrace		-> "}"
	TEquals		-> "="
	TBackslash	-> "\\"
	TArrow		-> "->"
	TIf		-> "if"
	TThen		-> "then"
	TElse		-> "else"
	TLet		-> "let"
	TIn		-> "in"
	TCase		-> "case"
	TOf		-> "of"
	TWhere		-> "where"
	TLParen		-> "("
	TRParen		-> ")"
	TLBracket	-> "["
	TRBracket	-> "]"
	TUnderscore	-> "_"
	TImpl		-> "=>"
	TPropArrow	-> "==>"
	TSimpArrow	-> "<==>"
	TComma		-> ","
	TBar		-> "|"
	TLArrow		-> "<-"
	TAnn _		-> "::"
	TEAnn _		-> ":::"
	TAsterisk	-> "*"
	TTrue		-> "True"
	TFalse		-> "False"
	TQuery		-> "::?"

	TLayoutBlock int    -> "LAYOUT{"++show int++"}"
	TLayoutFirst int    -> "LAYOUT<"++show int++">"
	Teof		    -> "EOF"


idStr  = fst
srcPos = snd
