--------------------------------------------------------------------------------
--
-- Copyright (C) 2002 The Chameleon Team
--
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License as published by the Free 
-- Software Foundation; either version 2 of the License, or (at your option) 
-- any later version. This program is distributed in the hope that it will be 
-- useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
-- See the GNU General Public License for more details. 
-- You should have received a copy of the GNU General Public License along
-- with this program; if not, write to the Free Software Foundation, Inc., 
-- 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
--
--------------------------------------------------------------------------------


-- contains definitions for, and functions over, user constraints
-- (originally in TD_Chr)

module UCons
where

import Types
import Terminal
import Pretty
import Justify
import Name
import Char
import List
import Misc (E(..), changeN)

import IOExts

-- User constraints:
-- Predicate symbols are represented by Int's
-- Index allows to identify user constraints,
-- necessary to prevent infinite prop rule application
-- (not we don't need to index builtins, assumption is that
-- builtins only appear in guards)


type Index = Integer
type PredSym = Name
data UCons = UC Index PredSym [Type] Just  deriving Eq


-- builds an unjustified UC
njUC i p ts = UC i p ts noJust

ucIx    (UC _ _ _ ix) = ix
ucTypes (UC _ _ ts _) = ts
ucJust  = ucIx
ucName (UC _ p _ _) = p
ucNameStr (UC _ p _ _) = nameStr p
newUCJust j (UC i p t _) = UC i p t j
addUCJust j (UC i p t k) = UC i p t (addJust k j)
catUCJust j (UC i p t k) = UC i p t (catJust k j)
newUCTypes t (UC i p _ j) = UC i p t j
newUCIx i (UC _ p t j) = UC i p t j
newUCName p (UC i _ t k) = UC i p t k

nubUCJust (UC i p t j) = UC i p t (nubJust j)

changeUCTypeN n t (UC i p ts j) = UC i p (changeN n ts t) j

isAdminUCons (UC i _ _ _) = i < 0

uconsPSym (UC _ psym _ _) = psym
uconsType (UC _ _ ts _)   = case ts of [t] -> t 
				       _   -> error "uconsType" 

instance Justifiable UCons where
    unJust (UC _ _ _ j) = unJust j
    reJust j (UC i p t _) = UC i p t j


instance PPrint UCons where
    pprint (UC ix name ts j) = do dts <- catListWith "," ts
			          return (text (show name) <> 
					  text "(" <> dts <> text ")") -- _" <>
--					    text (show j))

instance Show UCons where
  show (UC ix name tys j) = show name ++ "(" ++ commaList tys ++ ")" -- _" ++ show j


instance TC UCons where
   apply s (UC i p ts x) = UC i p (apply s ts) x
   fv (UC i p ts x) = (fv ts)
   occurs v (UC i p ts _) = occurs v ts
   rename (UC i p ts x) = do ts' <- rename ts
                             return (UC i p ts' x)
   format (UC i p ts _) = (show p) ++ "(" ++ (f ts)
     where
       f [] = ")"
       f [x] = (format x) ++ ")"
       f (x:xs) = (format x) ++ "," ++ (f xs)


instance Ord UCons where
   (<) (UC _ p1 ts1 _) (UC _ p2 ts2 _)
	= let comp [] _ = True    -- need my own lifted version on lists
	      comp _ [] = False
	      comp (x:xs) (y:ys)
                    | x < y = True
                    | x == y = comp xs ys
                    | otherwise = False
          in (p1 < p2) || (p1 == p2 && (comp ts1 ts2))


{- OLD VERSION, has some problems 
-- matchCons vs uc1 uc2
-- check whether there exists renaming function phi ( dom(phi) cap vs = emptyset)
-- such that uc1 = phi uc2
matchCons :: [TVar] -> [UCons] -> [UCons] -> Bool
matchCons vs uc1 uc2 = let uc1' = trace ("\n vs: " ++ (show vs) ++ "\n uc1: " ++ (show uc1) ++ "\n uc2: " ++ (show uc2)) uc1
                       in matchCons'' vs uc1' uc2

matchCons'' vs uc1 uc2 = case (matchCons2 vs idSubst (prepare uc1) (prepare uc2)) of
                         Succ _ -> True
                         Err t  -> trace ("matchCons: `" ++ t ++ "'\n") False -- error  t 
  where
    prepare = msort . removedups2 
    matchCons2 _ s [] [] = return s
    matchCons2 vs s ((u1@(UC i1 p1 ts1 _)):uc1) ((u2@(UC i2 p2 ts2 _)):uc2) =
         if p1 == p2 
         then do s' <- matchTypes2 vs s ts1 ts2
                 --matchCons2 vs s' uc1 uc2
                 case (matchCons2 vs s' uc1 uc2) of
                  Succ s -> Succ s
                  _      -> case (matchCons2 vs s' uc1 (u2:uc2)) of
                             Succ s -> Succ s
                             _      -> case (matchCons2 vs s' (u1:uc1) uc2) of
                                         Succ s -> Succ s
                                         Err t  -> Err t
              -- SET SEMANTICS
              -- matchCons [Eq a] [Eq a, Eq b] => True
              {- Therefore, in case of matchCons (u1:uc1) (u2:uc2)
                 and u1 matches u2, we check for
                     matchCons uc1 uc2  ||
                     matchCons (u1:uc1) uc2 ||
                     matchCons uc1 (u2:uc2)
              -}

                 
         else Err "no matching (1)"
    matchCons2 _ _ _ _ = Err "no matching (2)"
-}


-- matchCons vs uc1 uc2
-- check whether there exists renaming function phi 
-- ( dom(phi) cap vs = emptyset)
-- such that uc1 = phi uc2
-- returns the matching 
matchCons :: [TVar] -> [UCons] -> [UCons] -> (Bool,[(UCons,UCons)])
matchCons vs uc1 uc2 =  

		       let (vs',uc1',uc2') = 
{-			    trace ("\n vs="++ (show vs) ++ 
				   "\n uc1="++ (show uc1) ++ 
				   "\n uc2="++ (show uc2)) -}  (vs, uc1, uc2)
			   res = matchCons'' vs' uc1' uc2'
                       in  {- trace ("\n res=" ++ show res) -} res 
		       -- matchCons'' vs uc1 uc2


matchCons'' vs uc1 uc2 = check (permute (prepare uc2))
                               (\uc2 -> matchCons' vs (prepare uc1) uc2)

  where 
    prepare = msort . removedups2 
    permute [] = [[]] 
    permute (x:xs) = [ take i ys ++ [x] ++ drop i ys 
                                | ys <- permute xs, i <- [0..length ys] ] 
    check [] _ = (False,[])
    check (x:xs) f = let (r,m) = f x
                         muc1 = map fst m
                         muc1' = prepare muc1
                         r' = {- trace ("\n\tx: " ++ show x ++
				     "\n\txs: " ++ show xs ++
				     "\n\tr: " ++ show r ++
				     "\n\tmuc1': " ++ show muc1' ++
				     "\n\tprepare uc1: " ++ show (prepare uc1))$
				   -} (r && (muc1' == (prepare uc1)))
                     in if r' then (True,m)
                        else check xs f



-- matchCons' vs uc1 uc2
-- check whether there exists renaming function phi ( dom(phi) cap vs = emptyset)
-- such that uc1 = phi uc2
-- returns list of matches
-- order among uc2 remains fixed
matchCons' :: [TVar] -> [UCons] -> [UCons] -> (Bool,[(UCons,UCons)])
matchCons' vs uc1 uc2 = case (matchCons2 vs idSubst uc1 uc2 []) of
                         Succ (_,m) -> (True,m)
                         Err t  -> (False, [])
  where
    matchCons2 _ s _ [] m = return (s,m)
    matchCons2 vs s ((u1@(UC i1 p1 ts1 _)):uc1) ((u2@(UC i2 p2 ts2 _)):uc2) m =
      if p1 == p2 
      then let q = (matchTypes2 vs s ts1 ts2)
	   in case {- trace ("matchTypes2 " ++ show ts1 ++ " " ++ show ts2 ++ 
		       "\t: " ++ show q) -} q of
               Succ s' ->
                 --matchCons2 vs s' uc1 uc2
                 case (matchCons2 vs s' (uc1++[u1]) uc2 (m++[(u1,u2)])) of
                  Succ (s,m') -> Succ (s,m')
                  _      -> case (matchCons2 vs s' (u1:uc1) uc2 (m++[(u1,u2)])) of
                              Succ (s,m') -> Succ (s,m')
                              Err t  -> Err t
               Err t -> Err t
      else Err "no matching (1)"
    matchCons2 _ _ _ _ _ = Err "no matching (2)"


-- check wether uc1 is a `subset' of uc2
subsetCons :: [UCons] -> [UCons] -> Bool
subsetCons uc1 uc2 = (removedups2 (uc1++uc2)) == (removedups2 uc2)
      -- note == takes into account indices
      -- however all duplicates (indices don't count!) are removed
      -- therefore == works


-- unifyCons vs uc1 uc2
-- find phi such that dom(phi) cap vs = emptyset and phi uc1 = phi uc2
unifyCons :: [TVar] -> [UCons] -> [UCons] -> Bool
unifyCons vs uc1 uc2 = case (unifyCons2 vs idSubst (prepare uc1) (prepare uc2)) of
                         Succ _ -> True
                         Err  _ -> False
 where 
   prepare = removedups2 . sort    -- we put user-constraints in order
   unifyCons2 _ s [] [] = return s
   unifyCons2 vs s ((UC _ p1 ts1 _):uc1) ((UC _ p2 ts2 _):uc2) =
        if p1 == p2
        then if length ts1 == length ts2          -- check that same number of arguments
             then do s' <- unifyTypes vs s (zip ts1 ts2)   -- call proper unify on types function
                     unifyCons2 vs s' uc1 uc2
             else Err "unifyCons2: different arity"
        else Err "unifyCons2: not unifiable"
   unifyCons2 _ _ _ _ = Err "unifyCons2: not unifiable"


removedups2 [] = []
removedups2 (x:xs)
      | uelem x xs = removedups2 xs
      | otherwise = x:(removedups2 xs)

-- as above, but considers justifications as well
removedups3 [] = []
removedups3 (x:xs) | uelem x xs = removedups3 xs
		   | otherwise  = x : removedups3 xs
    where uelem _ [] = False
	  uelem (uc@(UC _ p1 ts1 j1)) ((UC _ p2 ts2 j2):xs)
	    | p1==p2 && ts1==ts2 && j1 == j2 = True
	    | otherwise                      = uelem uc xs


{-
showCommaList []	= ""
showCommaList [x]	= show x
showCommaList (x:xs)	= show x ++ ", " ++ showCommaList xs
-}
{-
showAnnCommaList []	= ""
showAnnCommaList [x]	= showAnn x
showAnnCommaList (x:xs)	= showAnn x ++ ", " ++ showAnnCommaList xs
-}


msort [] = []
msort (x:xs) = minsert x (msort xs)
 where
 minsert x [] = [x]
 minsert x (y:ys)
   | x < y = x:y:ys
   | otherwise = y:(minsert x ys)


-- we remove duplicates, 
-- indices don't count
uelem _ [] = False
uelem (uc@(UC _ p1 ts1 _)) ((UC _ p2 ts2 _):xs)
  | p1==p2 && ts1==ts2 = True
  | otherwise          = uelem uc xs

